"""
Provides decompiler integration by leveraging decomp2dbg
(https://github.com/mahaloz/decomp2dbg).

Communicates with the decomp2dbg decompiler plugins by
following the API laid out in https://github.com/mahaloz/decomp2dbg/blob/main/decompilers/server_template.py.

Code used as reference:
+ https://github.com/mahaloz/decomp2dbg/blob/main/decomp2dbg/clients/client.py
+ https://github.com/mahaloz/decomp2dbg/blob/main/decomp2dbg/clients/gdb/gdb_client.py
"""

from __future__ import annotations

import bisect
import os
import re
import tempfile
import xmlrpc
import xmlrpc.client
from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import cast

import niche_elf

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.vmmap
import pwndbg.color.syntax_highlight
import pwndbg.dbg_mod
import pwndbg.lib.cache
import pwndbg.lib.pretty_print as pretty_print
from pwndbg.color import message

# Note that XML RPC cannot send 64-bit ints (it is capped at 32 bits).
# We hope that rebased integers will never be more than 32-bits. If need be,
# we may send them as strings.


@dataclass
class GlobalVariable:
    # FIXME: Currently, global variables don't acknowledge their actual size.
    name: str
    # Mapped address in the address space.
    addr: int


@dataclass
class GlobalVariables:
    # The list is sorted by addr
    vars: list[GlobalVariable]


@dataclass
class FunctionHeader:
    name: str
    # Mapped address in the address space.
    addr: int
    size: int


@dataclass
class FunctionHeaders:
    # The list is sorted by addr
    funcs: list[FunctionHeader]


@dataclass
class RegisterVariable:
    name: str
    type: str
    reg_name: str


@dataclass
class StackVariable:
    name: str
    # Could easily be a type that the debugger doesn't know about.
    # E.g. Ida's __int64 and other MSVCisms. Can also be something
    # non-obvious like "void (*)()".
    type: str
    # == One of the the following two offsets is guaranteed to be non-None
    # The decompiler plugin code regarding this is touchy, may not always be
    # valid.
    # The offset of the variable's address from the stack pointer. Positive number.
    from_sp: int | None
    # The offset of the variable's address to the beginning of the stack frame
    # (which usually contains the saved return address). Positive number.
    from_frame: int | None


@dataclass
class RebasedStackVariable:
    name: str
    type: str
    # Actual valid address in the address space.
    addr: int


@dataclass
class FuncVariables:
    stack_vars: list[StackVariable]
    reg_vars: list[RegisterVariable]


@dataclass
class RebasedFuncVariables:
    stack_vars: list[RebasedStackVariable]
    reg_vars: list[RegisterVariable]


@dataclass
class FuncDecompilationResult:
    # The text containing the whole function decompilation.
    # Each element of the list is one line.
    # (contains the function signature and even stuff like IDA's
    # "// positive sp value has been detected, the output may be wrong!"
    # before the function signature)
    decompilation: list[str]
    # Says which line the requested address is in.
    # 0-indexed starting from the first line of the function.
    curr_line: int
    # The function name (not the signature!)
    func_name: str


class DecompilerID(Enum):
    IDA = "IDA"
    BINARYNINJA = "Binary Ninja"
    GHIDRA = "Ghidra"
    ANGR = "angr"


_api_name_to_id = {
    "ida": DecompilerID.IDA,
    "binaryninja": DecompilerID.BINARYNINJA,
    "ghidra": DecompilerID.GHIDRA,
    "angr": DecompilerID.ANGR,
}


class Error(Enum):
    OK = "Ok."
    NO_CONNECTION = "Not connected to a decompiler"
    BINARY_NOT_LOADED = "Couldn't find binary in address space"
    DEBUGGER_NOT_SUPPORTED = "The debugger does not support this operation"
    NOT_ALIVE = "The process is not alive"
    NO_FRAME = "No stack frame found."


# If the user wants to override our automatic detection
manual_binary_address: int = -1
manual_binary_path: str = ""


class DecompilerConnection:
    """
    Allows communication with the decompiler.

    The lifecycle of this object is tied to the connection to the compiler.
    It is only constructed after a successful connection, and must not be used
    after the connection dies.

    You should expect every function here to be able to throw ConnectionRefusedError.
    """

    # I allow this object to be live even if the process isn't live because I want
    # people to be able to connect to the decompiler in their gdbinit.

    """The XML RPC server that is connected to the decompiler."""
    server: xmlrpc.client.ServerProxy

    """The (host filesystem) path of the binary loaded as reported by the decompiler.
    It can be both an executable and a shared library.
    May not be relevant if manual_binary_address or manual_binary_path are set."""
    binary_path: str

    """Version information about the decompiler we are connected to. See
    plugin server_template.py for the format.
    """
    versions: dict[str, str]

    """The address of the start of the binary in the live process address space.
    Has value -1 if the process is not live or if the binary is not loaded yet."""
    # I allow (process live, binary not loaded) because we may be syncing with the
    # decompilation of a shared library that hasn't loaded yet.
    _binary_base_addr: int

    def __init__(self, server: xmlrpc.client.ServerProxy):
        self.server = server

        self.binary_path = str(self.server.binary_path())
        self.versions = cast(dict[str, str], self.server.versions())
        self._binary_base_addr = -1

        self._find_binary_addr(print_failure=True)

    def _find_binary_addr(self, print_failure: bool = False) -> None:
        if manual_binary_address != -1:
            # The user hardcoded the binary base address via `di setbase`.
            self._binary_base_addr = manual_binary_address
            return

        try:
            inf = pwndbg.dbg.selected_inferior()
        except pwndbg.dbg_mod.NoInferior:
            return

        if not inf.alive():
            return

        if manual_binary_path != "":
            # The user overrode what the decompiler says via `di setpath`.
            path: str = manual_binary_path
        else:
            path = self.binary_path

        # Try to find the binary in the address space.
        start_addr: int | None = pwndbg.aglib.vmmap.named_region_start(path, exact_match=True)

        if start_addr is None:
            # Try harder! (likely we are remote debugging)
            start_addr = pwndbg.aglib.vmmap.named_region_start(path, exact_match=False)

            if start_addr is None:
                if print_failure:
                    basename: str = os.path.basename(self.binary_path)
                    print(
                        message.notice(
                            f'The decompiled program "{basename}" doesn\'t seem to be loaded.'
                            " We will keep an eye out for it.\n"
                        )
                        + "If you know that it is actually loaded, check out "
                        + message.hint("`di setpath --help`")
                        + " or "
                        + message.hint("`di setbase --help`")
                        + ".\n"
                    )
                return
            self._binary_base_addr = start_addr
        else:
            self._binary_base_addr = start_addr

    def addr_to_mapped(self, rel_addr: int) -> int:
        """
        Takes an address relative to the image/file base and
        returns the actual address in the process' address
        space.

        self.binary_base_addr must be valid before calling this.
        """
        # If self.binary_base_addr is valid, so is
        # self._binary_base_addr :)
        assert self._binary_base_addr != -1
        return rel_addr + self._binary_base_addr

    def addr_to_relative(self, mapped_addr: int) -> int:
        """
        Takes an address from the live process' address space and returns
        the relative offset from the the image/file base.

        self.binary_base_addr must be valid before calling this.

        Assumes that this address is actually in the self.binary_path
        image rather than somewhere else. If you don't want to check this
        beforehand (because of performance), you at the very least need to
        check that the value returned here doesn't exceed XML-RPC int limits.
        """
        # If self.binary_base_addr is valid, so is
        # self._binary_base_addr :)
        assert self._binary_base_addr != -1
        return mapped_addr - self._binary_base_addr

    @property
    def binary_base_addr(self) -> int:
        if self._binary_base_addr == -1:
            self._find_binary_addr(print_failure=False)

        return self._binary_base_addr

    # ================
    # Decompiler interface.
    # Conforms to this file:
    # https://github.com/mahaloz/decomp2dbg/blob/77affe9ec1725e42739cf653a40ee6320452fd78/decompilers/server_template.py#L14
    # But the return values are repacked a bit for nicer usage.

    def disconnect(self) -> None:
        """
        Disconnects from the XML RPC server.

        Delete this object after running this function.
        """
        # XML RPC is stateless, there is no "disconnect"
        self.binary_path = (
            "You are using a disconnected DecompilerConnection. This is a bug in Pwndbg."
        )
        self._binary_base_addr = -2

    def decompile(self, mapped_addr: int) -> FuncDecompilationResult | None:
        """
        See IntegrationManager.decompile() for the function description.
        """
        if self.binary_base_addr == -1:
            return None

        rel_addr = self.addr_to_relative(mapped_addr)

        if rel_addr < xmlrpc.client.MININT or rel_addr > xmlrpc.client.MAXINT:
            # The user probably provided an address outside of the mappings of the
            # binary being decompiled.
            return None

        answer: dict[str, Any] = cast(dict[str, Any], self.server.decompile(rel_addr))

        if answer["decompilation"] is None:
            # Assuming all the other fields are as well
            return None

        return FuncDecompilationResult(
            decompilation=answer["decompilation"],
            curr_line=answer["curr_line"],
            func_name=answer["func_name"],
        )

    def function_data(self, mapped_addr: int) -> FuncVariables | None:
        """
        See IntegrationManager.function_data() for the function description.
        """
        if self.binary_base_addr == -1:
            return None

        rel_addr = self.addr_to_relative(mapped_addr)

        if rel_addr < xmlrpc.client.MININT or rel_addr > xmlrpc.client.MAXINT:
            # The user probably provided an address outside of the mappings of the
            # binary being decompiled.
            return None

        answer: dict[str, Any] = cast(dict[str, Any], self.server.function_data(rel_addr))

        if answer["stack_vars"] is None:
            return None

        stack_vars: list[StackVariable] = []
        reg_vars: list[RegisterVariable] = []

        for svar in answer["stack_vars"]:
            name = svar["name"]
            type_ = svar["type"]
            # .get() is needed because of ghidra
            from_sp_str: str | None = svar.get("from_sp")
            from_frame_str: str | None = svar.get("from_frame")
            from_sp: int | None = int(from_sp_str, 0) if from_sp_str is not None else None
            from_frame: int | None = int(from_frame_str, 0) if from_frame_str is not None else None

            stack_vars.append(
                StackVariable(name=name, type=type_, from_sp=from_sp, from_frame=from_frame)
            )

        for rvar in answer["reg_vars"]:
            name = rvar["name"]
            type_ = rvar["type"]
            reg_name = rvar["reg_name"]
            reg_vars.append(RegisterVariable(name=name, type=type_, reg_name=reg_name))

        return FuncVariables(stack_vars=stack_vars, reg_vars=reg_vars)

    def function_headers(self) -> FunctionHeaders | None:
        """
        See IntegrationManager.function_headers() for the function description.
        """
        if self.binary_base_addr == -1:
            return None

        answer: dict[str, Any] = cast(dict[str, Any], self.server.function_headers())

        functions: list[FunctionHeader] = []

        for key, value in answer.items():
            name: str = value["name"]
            size_: int = value["size"]
            addr: int = self.addr_to_mapped(int(key, 0))
            functions.append(FunctionHeader(name=name, addr=addr, size=size_))

        functions = sorted(functions, key=lambda f: f.addr)
        return FunctionHeaders(funcs=functions)

    def global_vars(self) -> GlobalVariables | None:
        """
        See IntegrationManager.global_vars() for the function description.
        """
        if self.binary_base_addr == -1:
            return None

        answer: dict[str, Any] = cast(dict[str, Any], self.server.global_vars())

        variables: list[GlobalVariable] = []

        for key, value in answer.items():
            addr: int = self.addr_to_mapped(int(key, 0))
            name: str = value["name"]
            variables.append(GlobalVariable(name=name, addr=addr))

        variables = sorted(variables, key=lambda v: v.addr)
        return GlobalVariables(vars=variables)

    def structs(self):
        # return self.server.structs()
        raise NotImplementedError()

    def breakpoints(self):
        # return self.server.breakpoints()
        raise NotImplementedError()

    # .binary_path and .versions are properties rather than functions

    def focus_address(self, mapped_addr: int) -> bool | None:
        """
        See IntegrationManager.focus_address() for the function description.
        """

        if self.binary_base_addr == -1:
            return None

        rel_addr = self.addr_to_relative(mapped_addr)

        if rel_addr < xmlrpc.client.MININT or rel_addr > xmlrpc.client.MAXINT:
            # The user probably provided an address outside of the mappings of the
            # binary being decompiled.
            return None

        answer: bool = cast(bool, self.server.focus_address(rel_addr))
        return answer

    # ================


class IntegrationManager:
    """
    A singleton class that manages all integration-related stuff.

    We can connect to only one decompiler at a time, and acknowledge only
    one file that decompiler is decompiling.
    (Could be relaxed in the future! Especially the latter.)

    All functions except connect() and disconnect() are no-op if we aren't
    connected.
    """

    def __init__(self) -> None:
        # Our connection to the decompiler.
        self._connection: DecompilerConnection | None = None

        # The local caches, invalidated on disconnect/reconnect or user request.
        # They MUST return None if self.connection is None.
        self._function_headers: FunctionHeaders | None = None
        self._global_vars: GlobalVariables | None = None
        self._decompiler_id: DecompilerID | None = None
        self._function_data: dict[int, FuncVariables | None] = {}

        # FIXME: Should really be fixed on decompiler plugin side.
        # Need to maintain this, otherwise the Ghidra decompilation pane is
        # useless. https://github.com/mahaloz/decomp2dbg/issues/131
        # It's fine if this isn't cleared on manager cache invalidation methinks.
        self.__func_curr_line: dict[str, int] = {}

        # Need to maintain so we can remove-symbol-file
        self._latest_symbol_file_path: str = ""

    def invalidate_caches(self) -> None:
        self._function_headers = None
        self._global_vars = None
        self._decompiler_id = None
        self._function_data.clear()

    def connect(self, host: str, port: int) -> bool:
        """
        Connects to the remote decompiler.

        Always invalidates the previous connection. This manager
        saves the new connection internally only if it succeeds.

        Returns True if the connection succeeded, otherwise False.
        """
        # Disconnect from previous connection. We also invalidate the cache here.
        self.disconnect()

        # Create a decompiler server connection and test it
        try:
            server = xmlrpc.client.ServerProxy(f"http://{host}:{port}")
            server.ping()
            # Success!
            self._connection = DecompilerConnection(server)
            return True
        except xmlrpc.client.Fault:
            # It's possible that Ghidra is on the other side.
            pass
        except ConnectionRefusedError:
            # The server probably just isn't running
            pass

        # The connection could have failed because it's a Ghidra connection on endpoint d2d
        try:
            server = xmlrpc.client.ServerProxy(f"http://{host}:{port}").d2d
            server.ping()
            # Success!
            self._connection = DecompilerConnection(server)
            return True
        except (ConnectionRefusedError, AttributeError):
            # We could also silently catch xmlrpc.client.Fault
            # but those are usually genuine bugs, so we kinda want to know about them.
            pass

        # Failed to connect.
        return False

    def remove_symbols(self, inf: pwndbg.dbg_mod.Process | None = None) -> bool:
        """
        Remove the decompiler symbols that we added latest.

        Returns whether we suceeded. Resets self._latest_symbol_file_path regardless
        of success.

        FIXME: Only works for GDB :(
        """
        path: str = self._latest_symbol_file_path
        self._latest_symbol_file_path = ""

        if not path:
            return False

        try:
            if inf is None:
                inf = pwndbg.dbg.selected_inferior()
            # FIXME: Only implemented in GDB :(
            if pwndbg.dbg.name() == pwndbg.dbg_mod.DebuggerType.GDB:
                return inf.remove_symbol_file(path)
        except pwndbg.dbg_mod.NoInferior:
            pass

        return False

    def disconnect(self) -> None:
        # We don't want to keep the data from the previous session.
        # FIXME: Ideally, we should also delete the convenience variables.
        self.remove_symbols()

        self.invalidate_caches()

        if self._connection is not None:
            self._connection.disconnect()
            self._connection = None

    # ==== Setters ====

    def update_symbols(self) -> tuple[int, Error]:
        """
        Update global variables and functions in the debugger.

        This always invalidates the cache for global variables and
        function headers, and requests them from the plugin.

        Returns the amount of synced symbols and an error diagnostic.

        Possible error values:
            NO_CONNECTION
            BINARY_NOT_LOADED
            NOT_ALIVE
            DEBUGGER_NOT_SUPPORTED

        FIXME: Currently they are all 8 bytes in size.
        """
        # We need to bail even if we are connected, but the binary is not loaded into
        # the address space yet.
        if self._connection is None:
            return 0, Error.NO_CONNECTION

        if self._connection.binary_base_addr == -1:
            return 0, Error.BINARY_NOT_LOADED

        # Invalidate the two caches.
        self._function_headers = None
        self._global_vars = None

        if pwndbg.dbg.name() == pwndbg.dbg_mod.DebuggerType.LLDB:
            # Until we implement add_symbol_file for LLDB.
            return 0, Error.DEBUGGER_NOT_SUPPORTED

        try:
            inf: pwndbg.dbg_mod.Process = pwndbg.dbg.selected_inferior()
        except pwndbg.dbg_mod.NoInferior:
            # Should never happen because we already know that binary_base_address is set.
            return 0, Error.NOT_ALIVE

        if not inf.alive():
            # Should never happen because we already know that binary_base_address is set.
            return 0, Error.NOT_ALIVE

        # Remove old symbol file.
        # If we don't do this, the symbols will stack (run `info func` in GDB).
        self.remove_symbols(inf)

        global_vars: GlobalVariables | None = self.global_vars()
        func_headers: FunctionHeaders | None = self.function_headers()
        # (name, address)
        syms_to_add: list[tuple[str, int]] = []
        # To get rid of duplicates
        sym_name_set: set[str] = set()

        if func_headers is not None:
            for func in func_headers.funcs:
                syms_to_add.append((func.name, func.addr))
                sym_name_set.add(func.name)

        if global_vars is not None:
            for var in global_vars.vars:
                clean_name = re.sub(r"[^a-zA-Z0-9_]", "_", var.name)
                # never re-add globals with the same name as a func
                if clean_name in sym_name_set:
                    continue

                syms_to_add.append((clean_name, var.addr))
                sym_name_set.add(clean_name)

        if not syms_to_add:
            return 0, Error.OK

        _, elf_path = tempfile.mkstemp(prefix="symbols-", suffix=".elf")
        elf = niche_elf.ELFFile(self._connection.binary_base_addr)

        for sym_name, sym_addr in syms_to_add:
            elf.add_generic_symbol(sym_name, sym_addr)

        elf.write(elf_path)
        inf.add_symbol_file(elf_path, self._connection.binary_base_addr)
        self._latest_symbol_file_path = elf_path
        # Delete the file after GDB closes the file descriptor.
        os.unlink(elf_path)
        return len(syms_to_add), Error.OK

    def _clean_type_str(self, type_str: str) -> str:
        # FIXME:
        # 1. this is too aggressive
        # 2. if we start adding types to the debugger then it doesn't matter
        # Copied from: https://github.com/mahaloz/decomp2dbg/blob/b23f2e232625c6ebe11b86c065c27b95da16aa3b/decomp2dbg/clients/gdb/gdb_client.py#L95
        if "__" in type_str:
            type_str = type_str.replace("__", "")
            idx = type_str.find("[")
            if idx != -1:
                type_str = type_str[:idx] + "_t" + type_str[idx:]
            else:
                type_str += "_t"
        type_str = type_str.replace("unsigned ", "u")

        return type_str

    def _try_setting_conv_var_with_type(self, name: str, value: str, type: str) -> bool:
        """
        Try setting a convenience variable with a type. If it fails try with void* .
        If that fails as well, thats okay, you can't win them all.

        Return True if we succeeded, False otherwise.
        """
        try:
            pwndbg.dbg.set_convenience_var(name, value, type)
            return True
        except Exception:
            pass

        try:
            pwndbg.dbg.set_convenience_var(name, value, "void*")
            return True
        except Exception:
            pass

        return False

    def update_function_variables(self) -> tuple[int, Error]:
        """
        Update debugger convnience varibles based on the function variables in the currently
        selected frame.

        This always fully invalidates the cache for function variables and requests
        them from the plugin.

        Returns:
            The number of variables we successfully updated in the debugger.
            An error diagnostic (NO_CONNECTION or NO_FRAME).

        FIXME: Currently this kinda doesn't work if it runs while we are in the function
        prologue. We should ideally run it only when we enter new functions and are past
        their prologues.
        """
        if self._connection is None:
            return 0, Error.NO_CONNECTION

        # We could do some updates without having a valid selected frame by using pwndbg.aglib.regs.sp ,
        # but this probably complicates the code uneccessarily (see some previous commits in the PR).
        # I'm simply not sure when exactly can selected_frame() actually return None.
        frame: pwndbg.dbg_mod.Frame | None = pwndbg.dbg.selected_frame()
        if frame is None:
            return 0, Error.NO_FRAME

        # Invalidate this whole cache.
        # We could invalidate just for frame.pc() for the purposes of this function, but we want to invalidate
        # this whole cache *somewhere* so non-local queries don't have stale data, and this function is a nice
        # place to do it.
        self._function_data.clear()

        rebased_vars: RebasedFuncVariables | None = self.get_function_vars_rebased_from_frame(frame)
        if rebased_vars is None:
            return 0, Error.OK

        nupdated: int = 0

        for reg_var in rebased_vars.reg_vars:
            cleaned_type: str = self._clean_type_str(reg_var.type)
            ok = self._try_setting_conv_var_with_type(
                reg_var.name, f"${reg_var.reg_name}", cleaned_type
            )
            nupdated += 1 if ok else 0

        for stack_var in rebased_vars.stack_vars:
            # Pointer to the type.
            cleaned_type = f"{self._clean_type_str(stack_var.type)}*"
            ok = self._try_setting_conv_var_with_type(
                stack_var.name, hex(stack_var.addr), cleaned_type
            )
            nupdated += 1 if ok else 0

        return nupdated, Error.OK

    # ==== Getters ====
    # All getters are either cheap (no RPC) operations, or cached.
    # The caching may be until stop, or until the cache is invalidated.

    def is_connected(self) -> bool:
        """
        Are we connected to a decompiler?

        Lightweight check, use a ping for an actual check.
        """
        return self._connection is not None

    def decompiler_id(self) -> DecompilerID | None:
        """
        Which decompiler are we connected to?
        """
        if self._connection is not None and self._decompiler_id is None:
            self._decompiler_id = _api_name_to_id[self._connection.versions["name"]]
        return self._decompiler_id

    def decompiler_name(self) -> str:
        """
        If we are connected, will return the name of the decompiler we are connected to.

        If we are not connected, will return "???".
        """
        decompilerid: DecompilerID | None = self.decompiler_id()
        if decompilerid is not None:
            return decompilerid.value
        return "???"

    def version_string(self) -> str | None:
        """
        Get a string with version information about the decompiler environment.
        """
        if self._connection is None:
            return None

        versions = self._connection.versions

        name: str = _api_name_to_id[versions["name"]].value
        ver: str = versions["version"]

        res = f"{name}: {ver}"
        # Add all other auxiliary information, no matter what it is.
        for key, value in versions.items():
            if key == "name" or key == "version":
                continue
            res += f"\n{name} {key}: {value}"

        return res

    def get_function_vars_rebased_from_frame(
        self, frame: pwndbg.dbg_mod.Frame
    ) -> RebasedFuncVariables | None:
        """
        Get function variables for the passed frame. Stack variables will have valid addresses rather than offsets.

        frame.pc() will be used to ask the debugger for the valid function variables. Note that it is possible that
        the same function returns different sets of variables at different PC's.

        frame.sp() and frame.start() are used to rebase stack variables based on decompiler-returned offsets. Register
        variables are unmodified.

        The RPC call to the decompiler when asking for variables is cached, but the rebasing is not, ergo this function
        call is relatively expensive.

        Arguments:
            frame: The frame to use for fetching variables.
        """
        if self._connection is None:
            return None

        addr = frame.pc()

        raw_func_data: FuncVariables | None = self.function_data(addr)
        if raw_func_data is None:
            return None

        try:
            inf = pwndbg.dbg.selected_inferior()
        except pwndbg.dbg_mod.NoInferior:
            return None

        if not inf.alive():
            return None

        # Nothing to do for registers
        new_stack_vars: list[RebasedStackVariable] = []

        frame_sp: int = frame.sp()
        frame_start: int | None = frame.start()

        for stack_var in raw_func_data.stack_vars:
            from_sp: int | None = stack_var.from_sp
            from_frame: int | None = stack_var.from_frame

            # We do not account for a stack going upwards. If you have that you have bigger issues.

            if from_sp is not None:
                # We prefer sp-offseted variables because calculating their actual address is more likely
                # to work.
                var_addr = int(frame_sp) + from_sp
                new_stack_vars.append(
                    RebasedStackVariable(name=stack_var.name, type=stack_var.type, addr=var_addr)
                )
            elif from_frame is not None and frame_start is not None:
                var_addr = frame_start - from_frame
                new_stack_vars.append(
                    RebasedStackVariable(name=stack_var.name, type=stack_var.type, addr=var_addr)
                )

        return RebasedFuncVariables(reg_vars=raw_func_data.reg_vars, stack_vars=new_stack_vars)

    def get_stack_var_dict_from_frame(self, frame: pwndbg.dbg_mod.Frame) -> dict[int, str]:
        """
        Ask the decompiler for stack variable offsets in this frame and resolve
        each variable to an actual address.

        The RPC call to the decompiler when asking for variables is cached, but the rebasing is not, and dict
        creating is not, ergo this function call is relatively expensive.

        Returns:
            A dictionary that maps (stack variable address) -> (stack variable name)
            for all variables in the given frame.
        """

        # The function will take care of connection checking etc.
        rebased_func_data: RebasedFuncVariables | None = self.get_function_vars_rebased_from_frame(
            frame
        )
        if not rebased_func_data:
            return {}

        result: dict[int, str] = {}
        for stack_var in rebased_func_data.stack_vars:
            result[stack_var.addr] = stack_var.name

        return result

    # FIXME: The implementation of cache_until tells me I shouldn't return mutable types
    # should I make a read-only dict class?
    @pwndbg.lib.cache.cache_until("stop")
    def get_stack_var_dict_all(self) -> dict[int, str]:
        """
        Take all valid stack frames (from the whole backtrace), ask the decompiler to
        figure out where they are, and map them to their actual addresses.

        You must not modify the object you got from this function (because of caching).

        Returns:
            A dictionary that maps (stack variable address) -> (stack variable name)
            for all currently valid stack frames.
        """
        if self._connection is None:
            return {}

        thread = pwndbg.dbg.selected_thread()
        if thread is None:
            return {}

        result: dict[int, str] = {}

        with thread.bottom_frame() as bottom_frame:
            cur_frame = bottom_frame
            # Crawl up the stack
            while cur_frame is not None:
                cur_variables = self.get_stack_var_dict_from_frame(cur_frame)
                # Merge dictionaries
                result = result | cur_variables
                cur_frame = cur_frame.parent()

        return result

    def decompile_pretty(self, mapped_addr: int, nlines: int) -> list[str] | None:
        """
        Get the prettified decompilation of a function.

        The following things are done:
        + syntax highlighting
        + '►' indicator at the current line
        + trimmed to only return `nlines` lines (surrounding the mapped_addr) (returns all lines if nlines == -1)

        Returns a list of strings each representing one line of the decompilation.
        """
        if self._connection is None:
            return None

        func_decomp: FuncDecompilationResult | None = self.decompile_raw(mapped_addr)

        if func_decomp is None:
            return None

        # Logic similar to pwndbg.commands.context.get_filename_and_formatted_source().

        decomp: list[str] = []

        if pwndbg.config.syntax_highlight:
            highlighted = pwndbg.color.syntax_highlight.syntax_highlight(
                "\n".join(func_decomp.decompilation), f"decompiled_{func_decomp.func_name}.c"
            )
            decomp = highlighted.splitlines()
        else:
            decomp = func_decomp.decompilation

        curr_line = func_decomp.curr_line

        # Ghidra often has this for some reason
        if decomp[0] == "":
            # This will make the line numbers off-by-one with the decompiler, but imo
            # saving screen space is more important.
            decomp = decomp[1:]
            # could be 0 or -1
            curr_line = (curr_line - 1) if curr_line > 0 else curr_line

        # Ghidra may return -1 (https://github.com/mahaloz/decomp2dbg/issues/131)
        # Cache the curr_line for this function if its valid.
        if curr_line == -1:
            if func_decomp.func_name in self.__func_curr_line:
                curr_line = self.__func_curr_line[func_decomp.func_name]
            else:
                curr_line = 0
        else:
            self.__func_curr_line[func_decomp.func_name] = curr_line

        formatted_decomp = pretty_print.format_source(list(decomp), nlines, curr_line)
        return formatted_decomp

    @pwndbg.lib.cache.cache_until("stop")
    def symbol_at_address(self, mapped_addr: int) -> str | None:
        """
        Returns name of a symbol (function or global variable) at given address,
        or None if there is nothing there.

        FIXME: Currently, global variables don't acknowledge their actual size.
        FIXME2: After update_symbols() is updated to acknowledge symbol sizes, this will be obsolete.
        """
        if self._connection is None:
            return None

        global_vars: GlobalVariables | None = self.global_vars()
        func_headers: FunctionHeaders | None = self.function_headers()

        if func_headers:
            # Binary search since the array is guaranteed to be sorted.
            idx = bisect.bisect_right(func_headers.funcs, mapped_addr, key=lambda f: f.addr)
            if idx != 0:
                possible_match = func_headers.funcs[idx - 1]
                if possible_match.addr <= mapped_addr < possible_match.addr + possible_match.size:
                    return possible_match.name

        if global_vars:
            # Binary search since the array is guaranteed to be sorted.
            idx = bisect.bisect_right(global_vars.vars, mapped_addr, key=lambda v: v.addr)
            if idx != 0:
                possible_match = global_vars.vars[idx - 1]
                if possible_match.addr == mapped_addr:
                    return possible_match.name

        return None

    # == Direct passthrough to the connection ==

    def decompile_raw(self, mapped_addr: int) -> FuncDecompilationResult | None:
        """
        Returns the decompilation of the function which contains address `mapped_addr`.

        Generally you should use self.decompile_pretty().
        """
        # I intentionally didn't name this function just `decompile` to prevent
        # people from accidentally using it when there is a better alternative.
        if self._connection is not None:
            return self._connection.decompile(mapped_addr)
        return None

    def function_data(self, mapped_addr: int) -> FuncVariables | None:
        """
        Returns the variables of the function which contains address `mapped_addr`.

        The "offset" field of the stack variables is poorly defined.

        The register variables are quite best effort and do not actually take
        the asked for address into account. In other words, the output for these
        may be just plain wrong.

        Function arguments are included in these variables.
        """
        if self._connection is not None and mapped_addr not in self._function_data:
            # This is a wacky cache. We could cache per function-name,
            # but that is not semantically correct (e.g. DWARF spec, split variables..)
            # In fact, even caching per-address is not semantically correct per DWARF since
            # the variable's locations may depend arbitrarily on the values of various registers.
            # But oke, it's probably fine :)
            # I'm okay with allowing `= None` here because callers that want the actual most
            # recent value will always just invalidate this cache anyway.
            self._function_data[mapped_addr] = self._connection.function_data(mapped_addr)

        # Using .get() for the case when there is no connection and the key is not in the cache.
        return self._function_data.get(mapped_addr, None)

    def function_headers(self) -> FunctionHeaders | None:
        """
        Returns the name, address and size off all functions in the binary, sorted
        by address.
        """
        if self._connection is not None and self._function_headers is None:
            self._function_headers = self._connection.function_headers()
        return self._function_headers

    def global_vars(self) -> GlobalVariables | None:
        """
        Returns the name and address of all global variables in the binary, sorted
        by address.
        """
        if self._connection is not None and self._global_vars is None:
            self._global_vars = self._connection.global_vars()
        return self._global_vars

    def structs(self):
        raise NotImplementedError()

    def breakpoints(self):
        raise NotImplementedError()

    def focus_address(self, mapped_addr: int) -> bool:
        """
        Focus (jump to) this address in the decompiler.
        """
        if self._connection is not None:
            ans = self._connection.focus_address(mapped_addr)
            return ans if ans is not None else False
        return False


manager: IntegrationManager = IntegrationManager()



from __future__ import annotations

import dataclasses
import json
import string
import struct
import textwrap
from abc import ABC
from abc import abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from enum import IntEnum
from typing import Any
from typing import Literal
from typing import cast

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.elf
import pwndbg.aglib.file
import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.aglib.symbol
import pwndbg.color.memory
import pwndbg.dintegration
import pwndbg.hexdump
import pwndbg.lib.cache
from pwndbg.color import generateColorFunction
from pwndbg.color import message
from pwndbg.color import theme

line_width = pwndbg.config.add_param(
    "go-dump-line-width", 80, "the soft line width for go-dump pretty printing"
)
indent_amount = pwndbg.config.add_param(
    "go-dump-indent-amount", 4, "the indent amount for go-dump pretty printing"
)

debug_color = theme.add_color_param(
    "go-dump-debug",
    "blue",
    "color for 'go-dump' command's debug info when --debug is specified",
)


@pwndbg.lib.cache.cache_until("start", "stop", "objfile")
def word_size() -> int:
    """
    Gets the Go word size for the current architecture.

    Values taken from https://github.com/golang/go/blob/49cdf0c42e320dfed044baa551610f081eafb781/src/cmd/compile/internal/types2/sizes.go#L230-L249
    """
    return {
        "x86-64": 8,
        "i386": 4,
        # Go cannot target i8086
        "arm": 4,
        # Go cannot target armcm
        "aarch64": 8,
        "powerpc": 8,
        "sparc": 8,
        # Go cannot target rv32 (but gccgo can)
        "rv64": 8,
        "mips": pwndbg.aglib.arch.ptrsize,  # pwndbg uses mips for both 32 and 64 bit
        "loongarch64": 8,
        "s390x": 8,
    }[pwndbg.aglib.arch.name]


@pwndbg.lib.cache.cache_until("start", "stop", "objfile")
def max_align() -> int:
    """
    Gets the Go maximum alignment for the current architecture.

    Values taken from https://github.com/golang/go/blob/49cdf0c42e320dfed044baa551610f081eafb781/src/cmd/compile/internal/types2/sizes.go#L230-L249
    """
    # Note: gccgo has max alignment 8 for arm and mips, which is different from the default compiler
    # Currently, no attempt is made to support gccgo, but it may be worth pursuing in the future
    return {
        "x86-64": 8,
        "i386": 4,
        # Go cannot target i8086
        "arm": 4,
        # Go cannot target armcm
        "aarch64": 8,
        "powerpc": 8,
        "sparc": 8,
        # Go cannot target rv32 (but gccgo can)
        "rv64": 8,
        "mips": pwndbg.aglib.arch.ptrsize,  # pwndbg uses mips for both 32 and 64 bit
        "loongarch64": 8,
        "s390x": 8,
    }[pwndbg.aglib.arch.name]


def _align(offset: int, n: int) -> int:
    ret = offset + n - 1
    return ret - ret % n


def compute_offsets(fields: Iterable[tuple[int, int]]) -> list[int]:
    """
    Given a list of (size, alignment) for struct field types,
    returns a list of field offsets for the struct.
    The last element will be the offset of the struct's end (the struct size).

    Layout computation taken from src/go/types/sizes.go commit 1b4f1dc
    """
    cur = 0
    ret = []
    max_align = 1
    for s, a in fields:
        cur = _align(cur, a)
        ret.append(cur)
        cur += s
        max_align = max(max_align, a)
    cur = _align(cur, max_align)
    ret.append(cur)
    return ret


def compute_named_offsets(fields: Iterable[tuple[str, int, int]]) -> dict[str, int]:
    """
    Like compute_offsets, but takes in field names and returns a dictionary
    mapping field name to offset instead.

    Also maps in a special $size field with the size of the struct.
    """
    offsets = compute_offsets([f[1:] for f in fields])
    ret = dict(zip([f[0] for f in fields] + ["$size"], offsets))
    ret["$align"] = max(f[2] for f in fields)
    return ret


def _indent(s: str) -> str:
    return textwrap.indent(s, " " * int(indent_amount))


@dataclass(frozen=True)
class FormatOpts:
    int_hex: bool = False
    debug: bool = False
    pretty: bool = False
    float_decimals: int | None = None

    def fmt_int(self, val: int) -> str:
        if self.int_hex:
            return hex(val)
        return str(val)

    def fmt_float(self, val: float) -> str:
        if self.float_decimals is not None:
            return format(val, f".{self.float_decimals}f")
        return str(val)

    def fmt_str(self, val: str) -> str:
        return json.dumps(val)

    def fmt_bytes(self, val: bytes) -> str:
        try:
            return self.fmt_str(val.decode("utf8"))
        except UnicodeDecodeError:
            return repr(bytes(val))

    def fmt_debug(self, val: str, default: str = "") -> str:
        if self.debug:
            return generateColorFunction(debug_color)(val)
        return default

    def fmt_elems(self, elems: Iterable[str]) -> str:
        if not self.pretty:
            return ", ".join(elems)
        # store elems in a list to not consume the iterable
        elems = list(elems)
        joined = ", ".join(elems)
        if len(joined) <= int(line_width):
            return joined
        joined = ",\n".join(elems)
        return f"\n{_indent(joined)}\n"

    def fmt_ptr(self, val: int) -> str:
        return pwndbg.color.memory.get_address_and_symbol(
            val, pwndbg.dintegration.manager.get_stack_var_dict_all()
        )


@dataclass
class Type(ABC):
    meta: GoTypeMeta | None

    @abstractmethod
    def dump(self, addr: int, fmt: FormatOpts = FormatOpts()) -> str:
        """Dump a type from memory given an address and format."""
        pass

    @abstractmethod
    def size(self) -> int:
        """
        Returns the size of a type in bytes.

        Used for computing array and struct layouts.
        """
        pass

    @abstractmethod
    def align(self) -> int:
        """
        Returns the alignment of a type in bytes.

        Used for computing array and struct layouts.
        """
        pass

    @abstractmethod
    def get_typename(self) -> str:
        """
        Returns the typename of a type. Should be reparsable via _parse_ty.

        Also used to get the string representation.
        """
        pass

    def is_cyclic(self) -> bool:
        """
        Checks if a type is cyclic (contains references to itself), e.g. type a []a
        """

        return _cyclic_helper(self, set())

    def additional_metadata(self) -> list[str]:
        """
        Returns a list of lines of additional metadata to dump from the `go-type` command.
        """
        return []

    def __str__(self) -> str:
        return self.get_typename()


def _cyclic_helper(val: Any, seen: set[int]) -> bool:
    if isinstance(val, Type):
        k = id(val)
        if k in seen:
            return True
        seen.add(k)
        return any(_cyclic_helper(v, seen) for v in val.__dict__.values())
    if isinstance(val, (list, tuple)):
        return any(_cyclic_helper(v, seen) for v in val)
    return False


def load_uint(data: bytes, endian: Literal["little", "big"] | None = None) -> int:
    return int.from_bytes(data, endian or pwndbg.aglib.arch.endian)


def load_int(data: bytes) -> int:
    n = load_uint(data)
    wrap = 1 << (len(data) * 8 - 1)
    if n >= wrap:
        n -= wrap * 2
    return n


def load_float(data: bytes) -> float:
    endian = ">" if pwndbg.aglib.arch.endian == "big" else "<"
    if len(data) == 4:
        return struct.unpack(endian + "f", data)[0]
    if len(data) == 8:
        return struct.unpack(endian + "d", data)[0]
    raise ValueError("Invalid float length")


# only warn with a given message once per execution
@pwndbg.lib.cache.cache_until("start")
def emit_warning(msg: str):
    print(message.warn(msg))


@pwndbg.lib.cache.cache_until("objfile")
def get_elf() -> pwndbg.aglib.elf.ELFInfo | None:
    try:
        return pwndbg.aglib.elf.get_elf_info_rebased(
            pwndbg.aglib.file.get_proc_exe_file(), pwndbg.aglib.proc.binary_base_addr()
        )
    except OSError:
        return None


def read_buildversion(addr: int) -> str:
    """
    Reads a Go runtime.buildVersion string to extract the version.
    """
    word = word_size()
    version_ptr = load_uint(pwndbg.aglib.memory.read(addr, word))
    version_len = load_uint(pwndbg.aglib.memory.read(addr + word, word))
    return "" if version_len == 0 else pwndbg.aglib.memory.read(version_ptr, version_len).decode()


@pwndbg.lib.cache.cache_until("objfile")
def get_go_version() -> tuple[int, ...] | None:
    """
    Try to determine the Go version used to compile the binary.

    None can be returned if the version couldn't be inferred,
    at which point it's probably best to assume latest version.
    """
    # if a runtime.buildVersion symbol exists, prefer that
    buildversion_addr = pwndbg.aglib.symbol.lookup_symbol_addr("runtime.buildVersion")
    if buildversion_addr is not None:
        version_string = read_buildversion(buildversion_addr)
    else:
        elf = get_elf()
        # could do a linear search through executable pages for "\xff Go buildinf:" as a fallback
        if elf is None:
            return None
        buildinfo = next(
            (cast(int, s["sh_addr"]) for s in elf.sections if s["x_name"] == ".go.buildinfo"),
            None,
        )
        # again, could do linear search
        if buildinfo is None:
            return None
        # check for flags & flagsVersionInl
        if (pwndbg.aglib.memory.read(buildinfo + 15, 1)[0] & 2) != 2:
            buildversion_addr = load_uint(pwndbg.aglib.memory.read(buildinfo + 16, word_size()))
            version_string = read_buildversion(buildversion_addr)
        else:
            version_string = read_varint_str(buildinfo + 32).decode()
    if version_string == "unknown":
        return None
    if not version_string.startswith("go"):
        emit_warning(f"Go version string {version_string!r} doesn't start with 'go'")
        return None

    # Cleanup a string that looks like "go1.25.5 X:nodwarf5"
    version_string = version_string.split(" ")[0]
    return tuple(int(x) for x in version_string[2:].split("."))


@pwndbg.lib.cache.cache_until("objfile")
def _get_moduledata_types() -> tuple[tuple[int, int], ...] | None:
    ret = []
    try:
        md = pwndbg.aglib.symbol.lookup_symbol("runtime.firstmoduledata")
        if md:
            while True:
                start = int(md["types"])
                end = int(md["etypes"])
                ret.append((start, end))
                if int(md["next"]):
                    md = md["next"].dereference()
                else:
                    return tuple(ret)
        else:
            emit_warning(
                "Warning: Could not find `runtime.firstmoduledata` symbol, so a heuristic is used instead"
            )
    except pwndbg.dbg_mod.Error as e:
        emit_warning(
            f"Warning: Exception '{e}' occurred while trying to parse `runtime.firstmoduledata`, so a heuristic is used instead"
        )
    return None


@pwndbg.lib.cache.cache_until("objfile")
def _guess_moduledata_types() -> int | None:
    # the type:* symbol can indicate type start
    type_start = pwndbg.aglib.symbol.lookup_symbol_addr("type:*")
    if type_start is not None:
        return type_start
    # otherwise, just assume that types are at the start of .rodata if there aren't any debug symbols
    # not a great workaround, but parsing moduledata manually is very version-dependent
    elf = get_elf()
    if elf is not None:
        addr = next(
            (cast(int, x["sh_addr"]) for x in elf.sections if x["x_name"] == ".rodata"),
            None,
        )
        return addr
    return None


def get_type_start(addr: int | None = None) -> int | None:
    """
    Given the address to a type, try to find the moduledata types section containing it.

    Necessary to determine the base address that the type name is offset by.
    """
    # try to get type start by traversing moduledata symbol
    # will only work if debug symbols are enabled
    md_types = _get_moduledata_types()
    if md_types is not None:
        for start, end in md_types:
            if addr is None or start <= addr < end:
                return start
        emit_warning(
            f"Warning: Type at {addr:#x} is out of bounds of all module data, so a heuristic is used instead"
        )
        # if we found at least one moduledata, use the start of the first one
        if md_types:
            return md_types[0][0]
    return _guess_moduledata_types()


def read_varint_str(addr: int) -> bytes:
    """
    Read a length-prefix string encoded with Go's variable length encoding.

    Implementation taken from https://github.com/golang/go/blob/9d33956503c0d96c0c5666d374173f7ac9756d98/src/internal/abi/type.go#L640-L649
    """
    orig_addr = addr
    strlen = 0
    while True:
        b = pwndbg.aglib.memory.read(addr, 1)[0]
        strlen = (strlen << 7) | (b & 0x7F)
        if b == 0x80 or strlen > 0x1000:
            # we're probably not actually reading a varint str and should just return some bytes to avoid infinite looping
            return pwndbg.aglib.memory.read(orig_addr, 16)
        addr += 1
        if not (b & 0x80):
            break
    # pwndbg.aglib.memory.read doesn't support 0-length reads
    if strlen == 0:
        return b""
    return pwndbg.aglib.memory.read(addr, strlen)


def read_type_name(addr: int) -> bytes:
    """
    Reads a Go type name given the address to the name.

    Go type names are stored as a 1 byte bitfield followed by a varint length prefixed string after 1.17.

    Prior to 1.17, they were stored as a 1 byte bitfield followed by a 2 byte length prefixed string.
    """
    vers = get_go_version()
    if vers is not None and vers < (1, 17):
        # reverse the bytestring because Go uses big endian
        strlen = load_uint(pwndbg.aglib.memory.read(addr + 1, 2), "big")
        if strlen == 0:
            return b""
        return pwndbg.aglib.memory.read(addr + 3, strlen)
    return read_varint_str(addr + 1)


class GoTypeKind(IntEnum):
    INVALID = 0
    BOOL = 1
    INT = 2
    INT8 = 3
    INT16 = 4
    INT32 = 5
    INT64 = 6
    UINT = 7
    UINT8 = 8
    UINT16 = 9
    UINT32 = 10
    UINT64 = 11
    UINTPTR = 12
    FLOAT32 = 13
    FLOAT64 = 14
    COMPLEX64 = 15
    COMPLEX128 = 16
    ARRAY = 17
    CHAN = 18
    FUNC = 19
    INTERFACE = 20
    MAP = 21
    POINTER = 22
    SLICE = 23
    STRING = 24
    STRUCT = 25
    UNSAFEPOINTER = 26

    def get_simple_name(self) -> str | None:
        # Gets the name of a simple type
        if self.BOOL <= self <= self.COMPLEX128:
            return self.name.lower()
        if self == self.STRING:
            return "string"
        if self == self.UNSAFEPOINTER:
            return "uintptr"
        return None


@dataclass
class GoTypeMeta:
    name: str
    kind: GoTypeKind
    addr: int
    size: int = 0
    align: int = 1
    direct_iface: bool = False


@dataclass
class BackrefType(Type):
    """
    A temporary placeholder type used when dumping recursive types, e.g. type a []a
    """

    key: int

    def dump(self, addr: int, fmt: FormatOpts = FormatOpts()):
        raise NotImplementedError(f"Cannot dump placeholder type {type(self).__name__}.")

    def size(self) -> int:
        raise NotImplementedError(
            f"Cannot get size of placeholder type {type(self).__name__}. Perhaps the type is ill-formed? (e.g. struct that contains itself without indirection)"
        )

    def align(self) -> int:
        raise NotImplementedError(
            f"Cannot get alignment of placeholder type {type(self).__name__}. Perhaps the type is ill-formed? (e.g. struct that contains itself without indirection)"
        )

    def get_typename(self) -> str:
        if self.meta:
            return f"runtime({self.meta.size}){self.meta.addr:#x}"
        return "..."


def decode_runtime_type(addr: int, keep_backrefs: bool = False) -> tuple[GoTypeMeta, Type | None]:
    """
    Decodes a runtime reflection type from memory, returning a (meta, type) tuplee.

    The layout assumed is as follows (taken from src/internal/abi/type.go commit 1b4f1dc):

    type Type struct {
        Size_       uintptr
        PtrBytes    uintptr
        Hash        uint32
        TFlag       TFlag
        Align_      uint8
        FieldAlign_ uint8
        Kind_       Kind
        Equal       func(unsafe.Pointer, unsafe.Pointer) bool
        GCData      *byte
        Str         NameOff
        PtrToThis   TypeOff
    }
    """

    cache: dict[int, tuple[GoTypeMeta, Type | None]] = {}
    (meta, rec_ty) = _inner_decode_runtime_type(addr, cache)

    if not keep_backrefs:
        rec_ty = _remove_backrefs(rec_ty, cache)
    return (meta, rec_ty)


def _remove_backrefs(ty: Any, cache: dict[int, tuple[GoTypeMeta, Type | None]]) -> Any:
    """
    Helper function to replace all _BackrefType instances after the cache is fully resolved.

    May mutate the argument.
    """
    if isinstance(ty, BackrefType):
        return cache[ty.key][1]
    if isinstance(ty, Type):
        d = ty.__dict__
        for k, v in d.items():
            d[k] = _remove_backrefs(v, cache)
        return ty
    if isinstance(ty, (list, tuple)):
        constructor = type(ty)
        return constructor(_remove_backrefs(x, cache) for x in ty)
    return ty


def _inner_decode_runtime_type(
    addr: int, cache: dict[int, tuple[GoTypeMeta, Type | None]]
) -> tuple[GoTypeMeta, Type | None]:
    """
    Internal function for decode_runtime_type with a cache to avoid recursive types.
    """

    if addr in cache:
        return cache[addr]
    word = word_size()
    offsets = compute_named_offsets(
        [
            ("Size_", word, word),  # uintptr
            ("PtrBytes", word, word),  # uintptr
            ("Hash", 4, 4),  # uint32
            ("TFlag", 1, 1),  # TFlag (alias for uint8)
            ("Align_", 1, 1),  # uint8
            ("FieldAlign_", 1, 1),  # uint8
            ("Kind_", 1, 1),  # Kind (alias for uint8)
            ("Equal", word, word),  # funcptr
            ("GCData", word, word),  # *byte
            ("Str", 4, 4),  # NameOff (alias for int32)
            ("PtrToThis", 4, 4),  # TypeOff (alias for int32)
        ]
    )
    load = lambda off, sz: load_uint(pwndbg.aglib.memory.read(addr + off, sz))
    type_start = get_type_start(addr)
    tflag = load(offsets["TFlag"], 1)
    if type_start is None:
        name = "unknown name"
    else:
        name_ptr = type_start + load(offsets["Str"], 4)
        bname = read_type_name(name_ptr)
        try:
            name = bname.decode()
        except UnicodeDecodeError:
            name = repr(bytes(bname))
    kind_raw = load(offsets["Kind_"], 1)
    # KindMask is set to (1 << 5) - 1
    try:
        kind = GoTypeKind(kind_raw & ((1 << 5) - 1))
    except ValueError:
        kind = GoTypeKind.INVALID
    if kind == GoTypeKind.INVALID:
        return (GoTypeMeta(f"invalid type `{name}` at {addr:#x}", kind, addr), None)
    # if TFlagExtraStar is set, remove the leading * from type name
    if (tflag & 2) and name.startswith("*"):
        name = name[1:]
    size = load(offsets["Size_"], word)
    align = load(offsets["Align_"], 1)
    meta = GoTypeMeta(
        name,
        kind,
        addr,
        size=size,
        align=align,
        direct_iface=(kind_raw & (1 << 5)) != 0,
    )
    cache[addr] = (meta, BackrefType(meta, addr))
    simple_name = kind.get_simple_name()

    def compute() -> tuple[GoTypeMeta, Type | None]:
        if simple_name is not None:
            return (meta, BasicType(meta, simple_name))
        if kind == GoTypeKind.FUNC:
            in_count = load(offsets["$size"], 2)
            out_count = load(offsets["$size"] + 2, 2)
            vararg_bit = 1 << 15
            is_vararg = bool(out_count & vararg_bit)
            out_count &= ~vararg_bit
            # functions store stuff after the uncommon type
            if tflag & 1:
                uncommon_type_size = 16  # nameoff, uint16, uint16, uint32, uint32
            else:
                uncommon_type_size = 0
            func_type_size = offsets["$size"] + 4
            # account for alignment when word size > 4
            if func_type_size % word != 0:
                func_type_size += word - func_type_size % word
            func_type_size += uncommon_type_size
            info = []
            for i in range(in_count + out_count):
                ty_ptr = load(func_type_size + i * word, word)
                (ty_meta, _) = _inner_decode_runtime_type(ty_ptr, cache)
                if i < in_count:
                    if is_vararg and i == in_count - 1:
                        suffix = "..."
                    else:
                        suffix = ""
                    info.append(f"Argument {i}{suffix}:")
                else:
                    info.append(f"Return value {i - in_count}:")
                info += [
                    f"    Type name: {ty_meta.name}",
                    f"    Type addr: {ty_ptr:#x}",
                ]
            return (meta, BasicType(meta, "funcptr", info))
        if kind == GoTypeKind.ARRAY:
            elem_ty_ptr = load(offsets["$size"], word)
            arr_len = load(offsets["$size"] + word * 2, word)
            elem_meta, elem_ty = _inner_decode_runtime_type(elem_ty_ptr, cache)
            # reserialize name to fix inconsistencies
            meta.name = f"[{arr_len}]{elem_meta.name}"
            return (meta, elem_ty and ArrayType(meta, elem_ty, arr_len))
        if kind == GoTypeKind.INTERFACE:
            methods_count = load(offsets["$size"] + word * 2, word)
            if methods_count == 0:
                return (meta, BasicType(meta, "any"))
            if type_start is None:
                return (
                    meta,
                    BasicType(meta, "interface", [f"Method count: {methods_count}"]),
                )
            info = []
            methods_ptr = load(offsets["$size"] + word, word)
            for i in range(methods_count):
                base = methods_ptr + i * 8
                meth_name_off = load_uint(pwndbg.aglib.memory.read(base, 4))
                inner_off = load_uint(pwndbg.aglib.memory.read(base + 4, 4))
                bmeth_name = read_type_name(type_start + meth_name_off)
                inner_ty_ptr = type_start + inner_off
                (inner_meta, _) = _inner_decode_runtime_type(inner_ty_ptr, cache)
                try:
                    meth_name = bmeth_name.decode()
                except UnicodeDecodeError:
                    meth_name = repr(bytes(bmeth_name))
                info += [
                    f"Method {meth_name}:",
                    f"    Type name: {inner_meta.name}",
                    f"    Type addr: {inner_ty_ptr:#x}",
                ]
            return (meta, BasicType(meta, "interface", info))
        if kind == GoTypeKind.MAP:
            key_ty_ptr = load(offsets["$size"], word)
            val_ty_ptr = load(offsets["$size"] + word, word)
            key_meta, key_ty = _inner_decode_runtime_type(key_ty_ptr, cache)
            if key_ty is None:
                return (meta, None)
            val_meta, val_ty = _inner_decode_runtime_type(val_ty_ptr, cache)
            if val_ty is None:
                return (meta, None)
            # reserialize name to fix inconsistencies
            meta.name = f"map[{key_meta.name}]{val_meta.name}"
            # Go maps are actually pointers, but the map here is not
            return (meta, PointerType(meta, MapType(meta, key_ty, val_ty)))
        if kind == GoTypeKind.POINTER:
            elem_ty_ptr = load(offsets["$size"], word)
            elem_meta, elem_ty = _inner_decode_runtime_type(elem_ty_ptr, cache)
            # reserialize name to fix inconsistencies
            meta.name = f"*{elem_meta.name}"
            return (meta, elem_ty and PointerType(meta, elem_ty))
        if kind == GoTypeKind.SLICE:
            elem_ty_ptr = load(offsets["$size"], word)
            elem_meta, elem_ty = _inner_decode_runtime_type(elem_ty_ptr, cache)
            # reserialize name to fix inconsistencies
            meta.name = f"[]{elem_meta.name}"
            return (meta, elem_ty and SliceType(meta, elem_ty))
        if kind == GoTypeKind.STRUCT:
            fields_ptr = load(offsets["$size"] + word, word)
            fields_count = load(offsets["$size"] + word * 2, word)
            fields: list[tuple[str, Type | str, int]] = []
            vers = get_go_version()
            if vers is not None and vers < (1, 19):
                offset_shift = 1
            else:
                offset_shift = 0
            for i in range(fields_count):
                base = fields_ptr + i * word * 3
                bfield_name = read_type_name(load_uint(pwndbg.aglib.memory.read(base, word)))
                try:
                    field_name = bfield_name.decode()
                except UnicodeDecodeError:
                    field_name = repr(bytes(bfield_name))
                field_ty_ptr = load_uint(pwndbg.aglib.memory.read(base + word, word))
                field_off = (
                    load_uint(pwndbg.aglib.memory.read(base + word * 2, word)) >> offset_shift
                )
                (field_meta, field_ty) = _inner_decode_runtime_type(field_ty_ptr, cache)
                if field_ty is None:
                    field_ty = field_meta.name
                fields.append((field_name, field_ty, field_off))
            fields.sort(key=lambda f: f[2])
            return (
                meta,
                StructType(meta, fields, size, align, None if name.startswith("struct ") else name),
            )
        # currently channels and functions are unsupported
        return (meta, None)

    ret = compute()
    if not isinstance(ret[1], BackrefType):
        cache[addr] = ret
    return ret


@dataclass
class BasicType(Type):
    """
    A primitive Go type.

    Complex numbers are laid out as a real and imaginary part (both floats).
    Strings are laid out as a pointer and a length.

    Methodless interfaces (the interface{} type) are denoted as any,
    and interfaces with methods are denoted as interface.

    Function pointers are denoted as funcptr.
    """

    name: str
    sz: int = dataclasses.field(init=False)
    algn: int = dataclasses.field(init=False)
    extra_meta: list[str] = dataclasses.field(default_factory=list)

    def dump(self, addr: int, fmt: FormatOpts = FormatOpts()) -> str:
        val = pwndbg.aglib.memory.read(addr, self.size())
        ty = self.name
        if ty == "byte":
            ty = "uint8"
        elif ty == "rune":
            ty = "int32"
        if ty in ("any", "interface"):
            word = word_size()
            if ty == "interface":
                iface_ptr = load_uint(val[:word])
                if iface_ptr == 0:
                    return "nil"
                ty_ptr = load_uint(pwndbg.aglib.memory.read(iface_ptr + word, word))
            else:
                ty_ptr = load_uint(val[:word])
            if ty_ptr == 0:
                return "nil"
            meta, parsed_inner = decode_runtime_type(ty_ptr)
            data_ptr = addr + word
            if not meta.direct_iface:
                data_ptr = load_uint(pwndbg.aglib.memory.read(data_ptr, word))
            prefix = fmt.fmt_debug(f"(ty @ {ty_ptr:#x}, data @ {data_ptr:#x}) ")
            if data_ptr == 0:
                return f"{prefix}({meta.name}) nil"
            if parsed_inner is not None:
                dump = parsed_inner.dump(data_ptr, fmt)
                return f"{prefix}({meta.name}) {dump}"
            return f"{prefix}({meta.name}) @ {data_ptr:#x}"
        if ty == "bool":
            return "true" if val != b"\x00" else "false"
        if ty == "funcptr":
            word = word_size()
            closure_addr = load_uint(val)
            f = load_uint(pwndbg.aglib.memory.read(closure_addr, word))
            return fmt.fmt_debug(f"(closure @ {closure_addr}) ") + fmt.fmt_ptr(f)
        if ty.startswith("int") or ty.startswith("uint"):
            if ty.startswith("int"):
                n = load_int(val)
            else:
                n = load_uint(val)
            if ty.endswith("ptr"):
                return fmt.fmt_ptr(n)
            return fmt.fmt_int(n)
        if ty.startswith("float"):
            return fmt.fmt_float(load_float(val))
        if ty.startswith("complex"):
            word = len(val) // 2
            real = fmt.fmt_float(load_float(val[:word]))
            im_val = load_float(val[word:])
            sign = "+"
            if im_val < 0:
                sign = "-"
                im_val = -im_val
            im = fmt.fmt_float(im_val)
            return f"({real} {sign} {im}i)"
        if ty == "string":
            word = word_size()
            ptr = load_uint(val[:word])
            strlen = load_uint(val[word:])
            # pwndbg.aglib.memory.read doesn't support 0-length reads
            if strlen == 0:
                data = b""
            else:
                data = pwndbg.aglib.memory.read(ptr, strlen)
            return fmt.fmt_debug(f"(str @ {ptr:#x}, len = {strlen}) ") + fmt.fmt_bytes(data)
        raise ValueError(f"Could not dump type {ty}.")

    def size(self) -> int:
        return self.sz

    def align(self) -> int:
        return min(self.algn, max_align())

    def get_typename(self) -> str:
        return self.name

    def additional_metadata(self) -> list[str]:
        return self.extra_meta

    def __post_init__(self) -> None:
        ty = self.name
        if ty in ("int8", "uint8", "bool", "byte"):
            self.sz = 1
            self.algn = 1
        elif ty in ("int16", "uint16"):
            self.sz = 2
            self.algn = 2
        elif ty in ("int32", "uint32", "float32", "rune"):
            self.sz = 4
            self.algn = 4
        elif ty in ("int64", "uint64", "float64", "complex64"):
            self.sz = 8
            self.algn = 8
        elif ty == "complex128":
            self.sz = 16
            self.algn = 8
        elif ty in ("int", "uint", "uintptr", "funcptr"):
            self.sz = word_size()
            self.algn = word_size()
        elif ty == "string":
            self.sz = word_size() * 2
            self.algn = word_size()
        elif ty in ("any", "interface"):
            self.sz = word_size() * 2
            self.algn = word_size()
        else:
            raise ValueError(
                f"Type {ty} is unknown. Use type hexdump[n] for an unknown type of size n."
            )


@dataclass
class SliceType(Type):
    """
    A slice type in Go, notated as []inner.

    Slices are laid out as a pointer, length, and capacity.
    """

    inner: Type

    def dump(self, addr: int, fmt: FormatOpts = FormatOpts()) -> str:
        word = word_size()
        val = pwndbg.aglib.memory.read(addr, word * 3)
        ptr = load_uint(val[:word])
        slice_len = load_uint(val[word : word * 2])
        ret = []
        for _ in range(slice_len):
            ret.append(fmt.fmt_debug(f"(elem @ {ptr:#x}) ") + self.inner.dump(ptr, fmt))
            ptr += self.inner.size()
        prefix = ""
        if fmt.debug:
            cap = load_uint(val[word * 2 :])
            prefix = fmt.fmt_debug(f"(cap = {cap}) ")
        return f"{prefix}[{fmt.fmt_elems(ret)}]"

    def size(self) -> int:
        return word_size() * 3

    def align(self) -> int:
        return word_size()

    def get_typename(self) -> str:
        return f"[]{self.inner}"

    def additional_metadata(self) -> list[str]:
        if self.inner.meta:
            return [
                f"Elem type name: {self.inner.meta.name}",
                f"Elem type addr: {self.inner.meta.addr:#x}",
            ]
        return []


@dataclass
class PointerType(Type):
    """
    A pointer type in Go, notated as *inner.
    """

    inner: Type

    def dump(self, addr: int, fmt: FormatOpts = FormatOpts()) -> str:
        word = word_size()
        ptr = load_uint(pwndbg.aglib.memory.read(addr, word))
        if ptr == 0:
            return "nil"
        inner = self.inner.dump(ptr, fmt)
        prefix = fmt.fmt_debug(f"(val @ {ptr:#x}) ")
        return f"{prefix}&{inner}"

    def size(self) -> int:
        return word_size()

    def align(self) -> int:
        return min(word_size(), max_align())

    def get_typename(self) -> str:
        return f"*{self.inner}"

    def additional_metadata(self) -> list[str]:
        # maps are returned as pointers to map in a parser
        # so show map metadata through the pointer metadata
        if isinstance(self.inner, MapType):
            info = self.inner.additional_metadata()
            if info:
                return info
        if self.inner.meta:
            return [
                f"Pointee type name: {self.inner.meta.name}",
                f"Pointee type addr: {self.inner.meta.addr:#x}",
            ]
        return []


@dataclass
class ArrayType(Type):
    """
    An array type in Go, notated as [count]inner.

    Arrays are laid out as contiguous data.
    """

    inner: Type
    count: int

    def dump(self, addr: int, fmt: FormatOpts = FormatOpts()) -> str:
        ret = []
        for _ in range(self.count):
            ret.append(fmt.fmt_debug(f"(elem @ {addr:#x}) ") + self.inner.dump(addr, fmt))
            addr += self.inner.size()
        return f"[{fmt.fmt_elems(ret)}]"

    def size(self) -> int:
        return self.inner.size() * self.count

    def align(self) -> int:
        return self.inner.align()

    def get_typename(self) -> str:
        return f"[{self.count}]{self.inner}"

    def additional_metadata(self) -> list[str]:
        if self.inner.meta:
            return [
                f"        Length: {self.count}",
                f"Elem type name: {self.inner.meta.name}",
                f"Elem type addr: {self.inner.meta.addr:#x}",
            ]
        return []


@dataclass
class MapType(Type):
    """
    A map type in Go, notated as map[key]val.

    Note that maps in Go are actually pointers to the inner map,
    but the map type printer here directly prints the inner map.

    Maps don't have a simple layout, and may reasonably change.
    """

    key: Type
    val: Type

    @staticmethod
    def is_swiss() -> bool:
        vers = get_go_version()
        return vers is None or vers >= (1, 24)

    @staticmethod
    def field_offsets_noswiss() -> dict[str, int]:
        """
        The layout for pre-1.24 maps is as follows (taken from src/runtime/map.go commit 1b4f1dc):

        type hmap struct {
            count      int
            flags      uint8
            B          uint8
            noverflow  uint16
            hash0      uint32
            buckets    unsafe.Pointer
            oldbuckets unsafe.Pointer
            nevacuate  uintptr
            extra      *mapextra
        }
        """
        word = word_size()
        offsets = compute_named_offsets(
            [
                ("count", word, word),  # int
                ("flags", 1, 1),  # uint8
                ("B", 1, 1),  # uint8
                ("noverflow", 2, 2),  # uint16
                ("hash0", 4, 4),  # uint32
                ("buckets", word, word),  # unsafe.Pointer
                ("oldbuckets", word, word),  # unsafe.Pointer
                ("nevacuate", word, word),  # uintptr
                ("extra", word, word),  # *mapextra
            ]
        )
        return offsets

    @staticmethod
    def field_offsets_swiss() -> dict[str, int]:
        """
        The layout for post-1.24 maps is as follows (taken from src/internal/runtime/map.go commit 4e63ae4):

        type Map struct {
            used              uint64
            seed              uintptr
            dirPtr            unsafe.Pointer
            dirLen            int
            globalDepth       uint8
            globalShift       uint8
            writing           uint8
            tombstonePossible bool
            clearSeq          uint64
        }
        """
        word = word_size()
        offsets = compute_named_offsets(
            [
                ("used", 8, 8),  # uint64
                ("seed", word, word),  # uintptr
                ("dirPtr", word, word),  # unsafe.Pointer
                ("dirLen", word, word),  # int
                ("globalDepth", 1, 1),  # uint8
                ("globalShift", 1, 1),  # uint8
                ("writing", 1, 1),  # uint8
                ("tombstonePossible", 1, 1),  # bool
                ("clearSeq", 8, 8),  # uint64
            ]
        )
        return offsets

    @staticmethod
    def field_offsets_swiss_inner() -> dict[str, int]:
        """
        The layout for the inner swissmap is as follows (taken from src/internal/runtime/maps/table.go commit 4e63ae4):

        type table struct {
            used       uint16
            capacity   uint16
            growthLeft uint16
            localDepth uint8
            index      int
            groups     groupsReference
        }
        """
        word = word_size()
        offsets = compute_named_offsets(
            [
                ("used", 2, 2),  # uint16
                ("capacity", 2, 2),  # uint16
                ("growthLeft", 2, 2),  # uint16
                ("localDepth", 1, 1),  # uint8
                ("index", word, word),  # uint8
                ("groups", 16, 8),  # groupsReference
            ]
        )
        return offsets

    @classmethod
    def field_offsets(cls) -> dict[str, int]:
        if cls.is_swiss():
            return cls.field_offsets_swiss()
        return cls.field_offsets_noswiss()

    @staticmethod
    def format_entries(
        entries: list[tuple[int, int, str, str]], fmt: FormatOpts = FormatOpts()
    ) -> str:
        # sort map by key, using integer comparison if possible
        try:
            entries.sort(key=lambda t: int(t[2], 0))
        except ValueError:
            entries.sort(key=lambda t: t[2])
        formatted = []
        for kp, vp, k, v in entries:
            prefix = fmt.fmt_debug(f"(key @ {kp:#x}, val @ {vp:#x}) ")
            formatted.append(f"{prefix}{k}: {v}")
        return f"{{{fmt.fmt_elems(formatted)}}}"

    def dump_noswiss(self, addr: int, fmt: FormatOpts = FormatOpts()) -> str:
        bucket_count = 8  # taken from src/internal/abi/map.go commit 1b4f1dc
        word = word_size()
        offsets = self.field_offsets_noswiss()
        val = pwndbg.aglib.memory.read(addr, offsets["$size"])
        load = lambda off, sz: load_uint(val[off : off + sz])
        num_buckets = 1 << load(offsets["B"], 1)
        num_oldbuckets = num_buckets >> 1
        oldbucket_base = load(offsets["oldbuckets"], word)
        bucket_base = load(offsets["buckets"], word)
        keysize = self.key.size()
        valsize = self.val.size()
        # technically need to worry about padding but every go arch has max alignment of 8 and bucket count is 8
        # so padding is never actually possible
        [tophash_start, keys_start, vals_start, overflow_start, bucket_size] = compute_offsets(
            [
                (bucket_count, 1),
                (keysize * bucket_count, 1),
                (valsize * bucket_count, 1),
                (word, word),
            ]
        )
        ret = []
        for i in range(num_buckets):
            bucket_ptr = bucket_base + bucket_size * i
            if oldbucket_base and i < num_oldbuckets:
                oldbucket_ptr = oldbucket_base + bucket_size * i
                oldbucket = pwndbg.aglib.memory.read(oldbucket_ptr, bucket_size)
                if not (1 < oldbucket[tophash_start] < 5):  # !evacuated(bucket)
                    bucket_ptr = oldbucket_ptr
            while bucket_ptr:
                bucket = pwndbg.aglib.memory.read(bucket_ptr, bucket_size)
                for j in range(bucket_count):
                    if bucket[tophash_start + j] > 1:  # !isEmpty(bucket.tophash[j])
                        key_ptr = bucket_ptr + keys_start + j * keysize
                        val_ptr = bucket_ptr + vals_start + j * valsize
                        k = self.key.dump(key_ptr, fmt)
                        v = self.val.dump(val_ptr, fmt)
                        ret.append((key_ptr, val_ptr, k, v))
                bucket_ptr = load_uint(bucket[overflow_start : overflow_start + word])
        return self.format_entries(ret, fmt)

    def dump_swiss(self, addr: int, fmt: FormatOpts = FormatOpts()) -> str:
        word = word_size()
        offsets = self.field_offsets_swiss()
        dir_ptr = load_uint(pwndbg.aglib.memory.read(addr + offsets["dirPtr"], word))
        dir_len = load_uint(pwndbg.aglib.memory.read(addr + offsets["dirLen"], word))
        slot_offsets = compute_named_offsets(
            [
                ("key", self.key.size(), self.key.align()),
                ("elem", self.val.size(), self.val.align()),
            ]
        )
        group_offsets = compute_named_offsets(
            [("ctrls", 8, 8), ("slots", slot_offsets["$size"] * 8, slot_offsets["$align"])]
        )
        if dir_len == 0:
            groups = [dir_ptr]
        else:
            table_ptrs = pwndbg.aglib.memory.read(dir_ptr, word * dir_len)
            tables = [load_uint(table_ptrs[i * word : i * word + word]) for i in range(dir_len)]
            table_offsets = self.field_offsets_swiss_inner()
            groups = []
            for table in tables:
                groups_ref = pwndbg.aglib.memory.read(table + table_offsets["groups"], 16)
                group_base = load_uint(groups_ref[:word])
                group_count = load_uint(groups_ref[8:]) + 1
                for _ in range(group_count):
                    groups.append(group_base)
                    group_base += group_offsets["$size"]
        ret = []
        for group_base in groups:
            group = pwndbg.aglib.memory.read(group_base, group_offsets["$size"])
            for i in range(8):
                if group[i] & 0x80 == 0:
                    off = 8 + slot_offsets["$size"] * i
                    key_ptr = group_base + off + slot_offsets["key"]
                    val_ptr = group_base + off + slot_offsets["elem"]
                    k = self.key.dump(key_ptr, fmt)
                    v = self.val.dump(val_ptr, fmt)
                    ret.append((key_ptr, val_ptr, k, v))
        return self.format_entries(ret, fmt)

    def dump(self, addr: int, fmt: FormatOpts = FormatOpts()) -> str:
        if self.is_swiss():
            return self.dump_swiss(addr, fmt)
        return self.dump_noswiss(addr, fmt)

    def size(self) -> int:
        return self.field_offsets()["$size"]

    def align(self) -> int:
        return min(8, max_align())

    def get_typename(self) -> str:
        return f"map[{self.key}]{self.val}"

    def additional_metadata(self) -> list[str]:
        ret = []
        if self.key.meta:
            ret += [
                f"Key type name: {self.key.meta.name}",
                f"Key type addr: {self.key.meta.addr:#x}",
            ]
        if self.val.meta:
            ret += [
                f"Val type name: {self.val.meta.name}",
                f"Val type addr: {self.val.meta.addr:#x}",
            ]
        return ret


@dataclass
class StructType(Type):
    """
    A struct type in Go, notated as struct(SIZE){FIELDS},
    where SIZE is the size of the struct in bytes,
    and FIELDS is a semicolon-separated list of OFFSET:NAME:TYPE fields.
    """

    fields: list[tuple[str, Type | str, int]]
    sz: int
    algn: int | None = None
    name: str | None = None

    def dump(self, addr: int, fmt: FormatOpts = FormatOpts()) -> str:
        vals = []
        for name, ty, off in self.fields:
            base = addr + off
            if isinstance(ty, str):
                vals.append((name, f"({ty}) @ {base:#x}"))
            else:
                vals.append((fmt.fmt_debug(f"(field @ {base:#x}) ") + name, ty.dump(base, fmt)))
        body = fmt.fmt_elems(f"{name}: {val}" for (name, val) in vals)
        name = self.name or "struct"
        return f"{name} {{{body}}}"

    def size(self) -> int:
        return self.sz

    def align(self) -> int:
        if self.algn is None:
            return max(
                (ty.align() for (_, ty, _) in self.fields if isinstance(ty, Type)), default=1
            )
        return self.algn

    def get_typename(self) -> str:
        body = ";".join(
            f"{off}:{name}:{ty}" for (name, ty, off) in self.fields if not isinstance(ty, str)
        )
        return f"struct({self.sz}){{{body}}}"

    def additional_metadata(self) -> list[str]:
        ret = []
        for name, ty, off in self.fields:
            if isinstance(ty, str) or not ty.meta:
                ret += [
                    f"Field {name}:",
                    f"    Offset: {off} ({off:#x})",
                    f"    Type: {ty}",
                ]
            else:
                ret += [
                    f"Field {name}:",
                    f"    Offset: {off} ({off:#x})",
                    f"    Type name: {ty.meta.name}",
                    f"    Type addr: {ty.meta.addr:#x}",
                ]
        return ret


@dataclass
class RuntimeType(Type):
    """
    A value of a runtime reflection type in Go, notated as runtime(SIZE)ADDRESS,
    where SIZE is the size of the type's value in bytes,
    and ADDRESS is the address of the type.

    This type is useful for serializing cyclic types.
    """

    sz: int
    addr: int
    algn: int | None = dataclasses.field(init=False, default=None)

    def dump(self, addr: int, fmt: FormatOpts = FormatOpts()) -> str:
        (meta, ty) = decode_runtime_type(self.addr)
        if ty is not None:
            return f"({meta.name}) {ty.dump(addr, fmt)}"
        return f"[error resolving type `{meta.name}` at {addr:#x}]"

    def size(self) -> int:
        return self.sz

    def align(self) -> int:
        if self.algn is None:
            (meta, _ty) = decode_runtime_type(self.addr)
            self.algn = meta.align
        return self.algn

    def get_typename(self) -> str:
        return f"runtime({self.sz}){self.addr:#x}"


_ident_first = set(string.ascii_letters + "_")
_ident_rest = _ident_first | set(string.digits)

# x is included for parsing purposes
hex_digits = set("0123456789abcdefABCDEFxX")


def _parse_posint(ty: str) -> tuple[int, str] | None:
    if not ty or ty[0] not in hex_digits:
        return None
    for i in range(1, len(ty)):
        if ty[i] not in hex_digits:
            break
    else:
        i = len(ty)
    try:
        return (int(ty[:i], 0), ty[i:])
    except ValueError:
        return None


def _parse_ident(ty: str) -> tuple[str, str] | None:
    if not ty or ty[0] not in _ident_first:
        return None
    for i in range(1, len(ty)):
        if ty[i] not in _ident_rest:
            break
    else:
        i = len(ty)
    return (ty[:i], ty[i:])


def _parse_basic_ty(ty: str) -> tuple[BasicType, str] | None:
    parse = _parse_ident(ty)
    if not parse:
        return None
    (ident, rest) = parse
    try:
        return (BasicType(None, ident), rest)
    except ValueError:
        if rest:
            return None
        # only raise an exception if it's a full string parse
        # otherwise the exception message could be inaccurate
        raise


def _parse_slice_ty(ty: str) -> tuple[SliceType, str] | None:
    if not ty.startswith("[]"):
        return None
    if (inner := _parse_type(ty[2:])) is None:
        return None
    return (SliceType(None, inner[0]), inner[1])


def _parse_pointer_ty(ty: str) -> tuple[PointerType, str] | None:
    if not ty.startswith("*"):
        return None
    if (inner := _parse_type(ty[1:])) is None:
        return None
    return (PointerType(None, inner[0]), inner[1])


def _parse_array_ty(ty: str) -> tuple[ArrayType, str] | None:
    if not ty.startswith("["):
        return None
    if (count := _parse_posint(ty[1:])) is None:
        return None
    if not count[1].startswith("]"):
        return None
    if (inner := _parse_type(count[1][1:])) is None:
        return None
    return (ArrayType(None, inner[0], count[0]), inner[1])


def _parse_map_ty(ty: str) -> tuple[MapType, str] | None:
    if not ty.startswith("map["):
        return None
    if (key := _parse_type(ty[4:])) is None:
        return None
    if not key[1].startswith("]"):
        return None
    if (val := _parse_type(key[1][1:])) is None:
        return None
    return (MapType(None, key[0], val[0]), val[1])


def _parse_struct_ty(ty: str) -> tuple[StructType, str] | None:
    if not ty.startswith("struct("):
        return None
    if (size_parse := _parse_posint(ty[7:])) is None:
        return None
    (size, cur) = size_parse
    if not cur.startswith("){"):
        return None
    cur = cur[2:]
    fields: list[tuple[str, Type | str, int]] = []
    is_first = True
    while cur:
        if cur.startswith("}"):
            return (StructType(None, fields, size), cur[1:])
        if is_first:
            is_first = False
        elif not cur.startswith(";"):
            return None
        cur = cur.lstrip(";")
        if (offset_parse := _parse_posint(cur)) is None:
            return None
        (field_offset, cur) = offset_parse
        if not cur.startswith(":"):
            return None
        if (name_parse := _parse_ident(cur[1:])) is None:
            return None
        (field_name, cur) = name_parse
        if not cur.startswith(":"):
            return None
        if (type_parse := _parse_type(cur[1:])) is None:
            return None
        (field_type, cur) = type_parse
        fields.append((field_name, field_type, field_offset))
    return None


def _parse_runtime_ty(ty: str) -> tuple[RuntimeType, str] | None:
    if not ty.startswith("runtime("):
        return None
    if (size_parse := _parse_posint(ty[8:])) is None:
        return None
    (size, rest) = size_parse
    if not rest.startswith(")"):
        return None
    if (addr_parse := _parse_posint(rest[1:])) is None:
        return None
    (addr, rest) = addr_parse
    return (RuntimeType(None, size, addr), rest)


def _parse_type(ty: str) -> tuple[Type, str] | None:
    for f in [
        _parse_runtime_ty,
        _parse_struct_ty,
        _parse_map_ty,
        _parse_array_ty,
        _parse_pointer_ty,
        _parse_slice_ty,
        _parse_basic_ty,
    ]:
        parse = f(ty)
        if parse is not None:
            return parse
    return None


def parse_type(ty: str) -> Type:
    ret = _parse_type(ty)
    if ret is None:
        raise ValueError(f"Type {ty} could not be parsed.")
    if ret[1]:
        raise ValueError(f"Type {ty} has trailing data.")
    return ret[0]



"""
Implements version and bugreport commands.
"""

from __future__ import annotations

import argparse
import os
import platform
import re
import sys
from subprocess import check_call
from subprocess import check_output
from tempfile import NamedTemporaryFile
from urllib.parse import quote

import pwndbg
import pwndbg.aglib
import pwndbg.commands
import pwndbg.dintegration
from pwndbg.color import message
from pwndbg.commands import CommandCategory
from pwndbg.dbg_mod import DebuggerType


def os_info():
    os_info = platform.system()

    if os_info.lower() == "linux" and os.path.isfile("/etc/os-release"):
        with open("/etc/os-release") as os_release:
            contents = os_release.read()
            match = re.search('PRETTY_NAME="?([^",\n]+)', contents)
            if match:
                os_info = match.group(1)

    return os_info


def module_version(module):
    try:
        return __import__(module).__version__
    except ImportError:
        return "not found"


def debugger_version():
    if pwndbg.dbg.is_gdblib_available():
        import gdb

        return f"GDB:      {gdb.VERSION}"
    return f"LLDB:     {'.'.join(map(str, pwndbg.dbg_mod.lldb.LLDB_VERSION))}"


def all_versions():
    py_version = sys.version.replace("\n", " ")
    most = (
        f"Pwndbg:   {pwndbg.__version__} ({os.uname().sysname})",
        f"Python:   {py_version}",
        debugger_version(),
        f"Capstone: {module_version('capstone')}",
        f"Unicorn:  {module_version('unicorn')}",
        f"Pwnlib:   {module_version('pwnlib')}",
    )

    integration_ver_text: str | None = pwndbg.dintegration.manager.version_string()
    if integration_ver_text is not None:
        return most + (integration_ver_text,)
    return most


def get_target_arch():
    arch_info = pwndbg.aglib.arch.name
    target = f"Target Arch: {arch_info}\n"

    if pwndbg.dbg.is_gdblib_available():
        import gdb

        # Note: this are only available if given arch is supported by GDB
        # (e.g., `gdb-multiarch` on Ubuntu)
        if arch_info in ("arm", "armcm", "aarch64"):
            arm_info = gdb.execute("show arm", to_string=True)
            target += f"ARM: {arm_info}\n"

        elif arch_info in ("mips", "mips64"):
            mips_info = gdb.execute("show mips", to_string=True)
            target += f"MIPS: {mips_info}\n"

    return target


def get_terminal_size():
    try:
        width_info = os.get_terminal_size().columns
        height_info = os.get_terminal_size().lines
    except OSError:
        # Terminal size may not be available in non-interactive environments (e.g., scripts, IDEs)
        width_info = height_info = "<unavailable>"

    return f"Terminal width: {width_info}, height: {height_info}\n"


def version_impl() -> None:
    """
    Implementation of the `version` command.
    """
    print("\n".join(map(message.system, all_versions())))


# In LLDB, this command is implemented as part of the Pwndbg CLI.
@pwndbg.commands.Command(
    "Displays Pwndbg and its important deps versions.",
    exclude_debuggers={DebuggerType.LLDB},
    category=CommandCategory.PWNDBG,
)
def version() -> None:
    version_impl()


bugreport_parser = argparse.ArgumentParser(description="Generate a bug report.")
bugreport_group = bugreport_parser.add_mutually_exclusive_group()
bugreport_group.add_argument(
    "--run-browser", "-b", action="store_true", help="Open browser on github/issues/new"
)
bugreport_group.add_argument(
    "--use-gh", "-g", action="store_true", help="Create issue using Github CLI"
)


@pwndbg.commands.Command(bugreport_parser, category=CommandCategory.PWNDBG)
def bugreport(run_browser=False, use_gh=False):
    ISSUE_TEMPLATE = """
<!--
Please see if the bug isn't reported already on: https://github.com/pwndbg/pwndbg/issues
and take part in discussion if it was.

Before reporting a new issue, make sure it happens on latest Pwndbg version.

Use the template below.
-->


### Description

<!--
Describe the problem you are having in a few paragraphs.
-->

### Steps to reproduce

<!--
What do we have to do to reproduce the problem?
If this is connected to particular C/asm code or a binary,
please provide the binary or if possible, a smallest C code that reproduces the issue.
-->

Session history:
```
{session_history}
```

### My setup

<!--
Show us your Pwndbg and any other relevant versions.
-->

```
{setup}
```"""
    setup = "\n".join(all_versions()) + "\n"
    setup += f"OS: {os_info()} ({platform.platform()}, {sys.byteorder} endian)\n"
    setup += f"OS ABI: {platform.uname().version}\n"
    setup += f"Charset: {sys.getdefaultencoding()}\n"
    setup += get_terminal_size()
    setup += get_target_arch()

    # Commented out for now: do we need this? It seems to be a bloat for GDB.
    # People rarely build custom GDB with fancy options...?
    # setup += get_debugger_configuration()

    session_history = get_debugger_session_history()

    issue_bugreport = ISSUE_TEMPLATE.format(setup=setup, session_history=session_history)
    print(issue_bugreport)

    please_please_submit = "Please submit the bugreport generated above at "
    github_issue_url = "https://github.com/pwndbg/pwndbg/issues/new"
    github_issue_body = "?body=" + quote(issue_bugreport)

    if use_gh:
        try:
            with NamedTemporaryFile("w", delete=True) as f:
                f.write(issue_bugreport)
                f.flush()
                check_call([os.environ.get("EDITOR", "vi"), f.name])
                check_call(["gh", "issue", "create", "--body-file", f.name])
        except Exception:
            print(please_please_submit + github_issue_url)
    elif run_browser:
        try:
            check_output(["xdg-open", github_issue_url + github_issue_body])
        except Exception:
            print(please_please_submit + github_issue_url)
    else:
        print(please_please_submit + github_issue_url)


def get_debugger_configuration():
    if pwndbg.dbg.is_gdblib_available():
        import gdb

        gdb_config = gdb.execute("show configuration", to_string=True).split("\n")
        return "\n" + "\n".join(gdb_config)

    # LLDB: TODO/FIXME: Do we need this?
    return ""


def get_debugger_session_history():
    if pwndbg.dbg.is_gdblib_available():
        import gdb

        # get saved history size (not including current gdb session)
        gdb_history_file = gdb.execute("show history filename", to_string=True)
        gdb_history_file = gdb_history_file[
            gdb_history_file.index('"') + 1 : gdb_history_file.rindex('"')
        ]
        gdb_history_len = 0
        try:
            with open(gdb_history_file) as f:
                gdb_history_len = len(f.readlines())
        except FileNotFoundError:
            pass

        max_command_no = 0
        history_commands = gdb.execute("show commands", to_string=True)
        if history_commands:
            history_commands = history_commands.split("\n")
            if len(history_commands) > 1:
                # The last element of the list is the `show commands` command we
                # just ran, so we need to get the second to last one
                last_command = history_commands[-2]
                max_command_no = int(last_command.split()[0]) - 1

        show_command_size = 10  # 'show command' returns 10 commands
        gdb_current_session_history = {}
        current_command_no = gdb_history_len + 1

        while current_command_no <= max_command_no:
            cmds = gdb.execute(
                "show commands " + str(current_command_no + (show_command_size // 2) + 1),
                to_string=True,
            ).split("\n")[:-1]
            for cmd in cmds:
                cmd_no, cmd = cmd.split(maxsplit=1)
                cmd_no = int(cmd_no)
                if cmd_no <= gdb_history_len:
                    continue
                if current_command_no > max_command_no:
                    break
                gdb_current_session_history[cmd_no] = cmd
                current_command_no += 1

        gdb_current_session_history = (v for (k, v) in sorted(gdb_current_session_history.items()))
        return "\n".join(gdb_current_session_history)

    # LLDB: TODO/FIXME: Not yet supported
    return "<session history not supported on lldb yet>"



from __future__ import annotations

import argparse

import pwndbg.aglib.shellcode
import pwndbg.commands
import pwndbg.dbg_mod
import pwndbg.lib.memory
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="""
Calls the mprotect syscall and prints its result value.

Note that the mprotect syscall may fail for various reasons
(see `man mprotect`) and a non-zero error return value
can be decoded with the `errno <value>` command.
""",
)
parser.add_argument(
    "addr", help="Page-aligned address to all mprotect on.", type=pwndbg.commands.sloppy_gdb_parse
)
parser.add_argument(
    "length",
    help="Count of bytes to call mprotect on. Needs to be multiple of page size.",
    type=int,
)
parser.add_argument(
    "prot", help='Prot string as in mprotect(2). Eg. "PROT_READ|PROT_EXEC", "rx", or "5"', type=str
)

SYS_MPROTECT = 0x7D

prot_dict = {
    "PROT_NONE": 0x0,
    "PROT_READ": 0x1,
    "PROT_WRITE": 0x2,
    "PROT_EXEC": 0x4,
}


def prot_str_to_val(protstr: str) -> int:
    """
    Converts a protection string to an integer. Formats include:
     - A positive integer, like 3
     - A combination of r, w, and x, like rw
     - A combination of PROT_READ, PROT_WRITE, and PROT_EXEC, like PROT_READ|PROT_WRITE
    """
    protstr = protstr.upper()
    if "PROT" in protstr:
        prot_int = 0
        for k, v in prot_dict.items():
            if k in protstr:
                prot_int |= v
        return prot_int
    if all(x in "RWX" for x in protstr):
        prot_int = 0
        for c in protstr:
            if c == "R":
                prot_int |= 1
            elif c == "W":
                prot_int |= 2
            elif c == "X":
                prot_int |= 4
        return prot_int
    try:
        return int(protstr, 0)
    except ValueError:
        raise ValueError("Invalid protection string passed into mprotect")


def prot_val_to_str(protval: int) -> str:
    if protval == 0:
        return "PROT_NONE"
    ret = []
    for k, v in prot_dict.items():
        if protval & v:
            ret.append(k)
    return "|".join(ret)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MEMORY,
    examples="""
mprotect $rsp 4096 PROT_READ|PROT_WRITE|PROT_EXEC
mprotect $rsp 4096 rwx
mprotect $rsp 4096 7
mprotect some_symbol 0x1000 PROT_NONE
""",
)
@pwndbg.commands.OnlyWhenRunning
def mprotect(addr, length, prot) -> None:
    prot_int = prot_str_to_val(prot)
    orig_addr = int(addr)
    aligned = pwndbg.lib.memory.page_align(orig_addr)

    async def ctrl(ec: pwndbg.dbg_mod.ExecutionController):
        print(
            f"calling mprotect on address {aligned:#x} with protection {prot_int} ({prot_val_to_str(prot_int)})"
        )

        ret = await pwndbg.aglib.shellcode.exec_syscall(
            ec, "SYS_mprotect", aligned, int(length) + orig_addr - aligned, int(prot_int)
        )
        print(f"mprotect returned {ret}")

    pwndbg.dbg.selected_inferior().dispatch_execution_controller(ctrl)



from __future__ import annotations

from collections.abc import Callable

# Emulator currently requires GDB, and we only use it here for type checking.
from typing import TYPE_CHECKING
from typing import Literal

from capstone6pwndbg import *  # noqa: F403
from capstone6pwndbg.arm import *  # noqa: F403
from pwnlib.util.misc import align_down
from typing_extensions import override

import pwndbg.aglib
import pwndbg.aglib.disasm.arch
import pwndbg.aglib.saved_register_frames
import pwndbg.lib.disasm.helpers as bit_math
from pwndbg.aglib.disasm.instruction import EnhancedOperand
from pwndbg.aglib.disasm.instruction import InstructionCondition
from pwndbg.aglib.disasm.instruction import PwndbgInstruction

if TYPE_CHECKING:
    from pwndbg.emu.emulator import Emulator

# Note: this map does not contain all the Arm32 shift types, just the ones relevent to register and memory modifier operations
ARM_BIT_SHIFT_MAP: dict[int, Callable[[int, int, int], int]] = {
    ARM_SFT_ASR: bit_math.arithmetic_shift_right,
    ARM_SFT_LSL: bit_math.logical_shift_left,
    ARM_SFT_LSR: bit_math.logical_shift_right,
    ARM_SFT_ROR: bit_math.rotate_right,
}

ARM_SINGLE_LOAD_INSTRUCTIONS = {
    ARM_INS_LDRB: 1,
    ARM_INS_LDRSB: -1,
    ARM_INS_LDRH: 2,
    ARM_INS_LDRSH: -2,
    ARM_INS_LDR: 4,
    ARM_INS_LDRBT: 1,
    ARM_INS_LDRSBT: -1,
    ARM_INS_LDRHT: 2,
    ARM_INS_LDRSHT: -2,
    ARM_INS_LDRT: 4,
    ARM_INS_LDREXB: 1,
    ARM_INS_LDREXH: 2,
    ARM_INS_LDREX: 4,
}

ARM_SINGLE_STORE_INSTRUCTIONS = {
    ARM_INS_STRB: 1,
    ARM_INS_STRH: 2,
    ARM_INS_STR: 4,
    ARM_INS_STRBT: 1,
    ARM_INS_STRHT: 2,
    ARM_INS_STRT: 4,
}

ARM_EXCLUSIVE_STORE_INSTRUCTIONS = {
    ARM_INS_STREXB: 1,
    ARM_INS_STREXH: 2,
    ARM_INS_STREX: 4,
}

ARM_MATH_INSTRUCTIONS = {
    ARM_INS_ADD: "+",
    ARM_INS_ADDW: "+",
    ARM_INS_SUB: "-",
    ARM_INS_ORR: "|",
    ARM_INS_AND: "&",
    ARM_INS_EOR: "^",
    ARM_INS_UDIV: "/",
    ARM_INS_SDIV: "/",
    ARM_INS_MUL: "*",
    ARM_INS_UMULL: "*",
    ARM_INS_SMULL: "*",
}

ARM_SHIFT_INSTRUCTIONS = {
    ARM_INS_ASR: ">>s",
    ARM_INS_ALIAS_ASR: ">>s",
    ARM_INS_LSR: ">>",
    ARM_INS_ALIAS_LSR: ">>",
    ARM_INS_LSL: "<<",
    ARM_INS_ALIAS_LSL: "<<",
    ARM_INS_ROR: ">>r",
    ARM_INS_ALIAS_ROR: ">>r",
}

# All of these instructions can write to the PC
# https://developer.arm.com/documentation/ddi0406/cb/Application-Level-Architecture/Application-Level-Programmers--Model/ARM-core-registers/Writing-to-the-PC?lang=en
# If they do write to PC, Capstone gives the instructions the `ARM_GRP_JUMP` group
# Note that we don't have the flag-setting variants - "ands", "subs" - because these generate an illegal instruction interrupt at runtime
ARM_CAN_WRITE_TO_PC_INSTRUCTIONS = {
    ARM_INS_LDM,
    ARM_INS_ALIAS_LDM,
    ARM_INS_POP,
    ARM_INS_ALIAS_POP,
    ARM_INS_LDR,
    ARM_INS_ADC,
    ARM_INS_ADD,
    ARM_INS_ADR,
    ARM_INS_AND,
    ARM_INS_ASR,
    ARM_INS_ALIAS_ASR,
    ARM_INS_BIC,
    ARM_INS_EOR,
    ARM_INS_LSL,
    ARM_INS_ALIAS_LSL,
    ARM_INS_LSR,
    ARM_INS_ALIAS_LSR,
    ARM_INS_MOV,
    ARM_INS_MVN,
    ARM_INS_ORR,
    ARM_INS_ROR,
    ARM_INS_ALIAS_ROR,
    ARM_INS_RRX,
    ARM_INS_ALIAS_RRX,
    ARM_INS_RSB,
    ARM_INS_RSC,
    ARM_INS_SBC,
    ARM_INS_SUB,
}


def itstate_from_cpsr(cpsr_value: int) -> int:
    """
    ITSTATE == If-Then execution state bits for the Thumb IT instruction
    The ITSTATE bits are spread across 3 sections of Arm flags register to a total of 8 bits.
    This function extracts them and reorders the bits into their logical order
    - https://developer.arm.com/documentation/ddi0403/d/System-Level-Architecture/System-Level-Programmers--Model/Registers/The-special-purpose-program-status-registers--xPSR#:~:text=shows%20the%20assignment%20of%20the%20ICI/IT%20bits.

    Bits of the flags register: EPSR[26:25]    EPSR[15:12]    EPSR[11:10]
    Bits of ITSTATE:            IT[1:0]        IT[7:4]        IT[3:2]

    The lower 5 bits has information that indicates the number of instructions in the IT Block.
    The top 3 bits indicate the base condition of the block.
    - https://developer.arm.com/documentation/ddi0406/cb/Application-Level-Architecture/Application-Level-Programmers--Model/Execution-state-registers/IT-block-state-register--ITSTATE?lang=en

    If the value is zero, it means we are not in an IT block.
    """

    return (
        ((cpsr_value >> 25) & 0b11)
        | ((cpsr_value >> 10) & 0b11) << 2
        | ((cpsr_value >> 12) & 0b1111) << 4
    )


# This class enhances both ARM A-profile and ARM M-profile (Cortex-M)
class ArmDisassemblyAssistant(pwndbg.aglib.disasm.arch.DisassemblyAssistant):
    def __init__(self, architecture, flags_reg: Literal["cpsr", "xpsr"]) -> None:
        super().__init__(architecture)

        self.flags_reg = flags_reg

        self.annotation_handlers: dict[int, Callable[[PwndbgInstruction, Emulator], None]] = {
            # MOV
            ARM_INS_MOV: self._common_move_annotator,
            ARM_INS_MOVW: self._common_move_annotator,
            # MOVT
            ARM_INS_MOVT: self._common_generic_register_destination,
            # MOVN
            ARM_INS_MVN: self._common_generic_register_destination,
            # CMP
            ARM_INS_CMP: self._common_cmp_annotator_builder(flags_reg, "-"),
            # CMN
            ARM_INS_CMN: self._common_cmp_annotator_builder(flags_reg, "+"),
            # TST (bitwise "and")
            ARM_INS_TST: self._common_cmp_annotator_builder(flags_reg, "&"),
            # TEQ (bitwise exclusive "or")
            ARM_INS_TEQ: self._common_cmp_annotator_builder(flags_reg, "^"),
        }

    @override
    def _set_annotation_string(self, instruction: PwndbgInstruction, emu: Emulator) -> None:
        if instruction.id in ARM_SINGLE_LOAD_INSTRUCTIONS:
            read_size = ARM_SINGLE_LOAD_INSTRUCTIONS[instruction.id]
            self._common_load_annotator(
                instruction,
                emu,
                instruction.operands[1].before_value,
                abs(read_size),
                read_size < 0,
                4,
                instruction.operands[0].str,
                instruction.operands[1].str,
            )
        elif instruction.id in ARM_SINGLE_STORE_INSTRUCTIONS:
            self._common_store_annotator(
                instruction,
                emu,
                instruction.operands[1].before_value,
                instruction.operands[0].before_value,
                ARM_SINGLE_STORE_INSTRUCTIONS[instruction.id],
                instruction.operands[1].str,
            )
        elif instruction.id in ARM_EXCLUSIVE_STORE_INSTRUCTIONS:
            # These store instructions include the "Store Register Exclusive", which
            # have an additional register at the front which pushes the source and destination one to the right.
            self._common_store_annotator(
                instruction,
                emu,
                instruction.operands[-1].before_value,
                instruction.operands[-2].before_value,
                ARM_EXCLUSIVE_STORE_INSTRUCTIONS[instruction.id],
                instruction.operands[-1].str,
            )
        elif instruction.id in ARM_MATH_INSTRUCTIONS:
            # In Arm assembly, if there are two operands, than the first source operand is also the destination
            # Example: add    sl, r3
            # Or, it can be a seperate register. We use -1 and -2 indexes here to access the source operands either way
            self._common_binary_op_annotator(
                instruction,
                emu,
                instruction.operands[0],
                instruction.operands[-2].before_value,
                instruction.operands[-1].before_value,
                ARM_MATH_INSTRUCTIONS[instruction.id],
            )
        elif instruction.id in ARM_SHIFT_INSTRUCTIONS:
            # The encoding of shifts has changed between past Capstone versions: https://github.com/capstone-engine/capstone/pull/2638
            # This check avoids a crash
            if len(instruction.operands) == 3:
                self._common_binary_op_annotator(
                    instruction,
                    emu,
                    instruction.operands[0],
                    instruction.operands[1].before_value_no_modifiers,
                    instruction.operands[2].before_value,
                    ARM_SHIFT_INSTRUCTIONS[instruction.id],
                )
        else:
            self.annotation_handlers.get(instruction.id, lambda *a: None)(instruction, emu)

    @override
    def _prepare(
        self, instruction: PwndbgInstruction, emu: pwndbg.aglib.disasm.arch.Emulator
    ) -> None:
        if CS_GRP_INT in instruction.groups:
            # https://github.com/capstone-engine/capstone/issues/2630
            instruction.groups.remove(CS_GRP_CALL)

        # Disable Unicorn while in IT instruction blocks since Unicorn cannot be paused in it.
        flags_value = pwndbg.aglib.regs.read_reg(self.flags_reg)
        it_state = itstate_from_cpsr(flags_value)

        if (instruction.id == ARM_INS_IT or it_state != 0) and emu:
            emu.valid = False

    @override
    def _condition(self, instruction: PwndbgInstruction, emu: Emulator) -> InstructionCondition:
        if ARM_GRP_JUMP in instruction.groups:
            if instruction.id in ARM_CAN_WRITE_TO_PC_INSTRUCTIONS:
                # Since Capstone V6, instructions that write to the PC are given the jump group.
                # However, in Pwndbg code, unless stated otherwise, jumps are assumed to be conditional, so we set this attribute
                # to indicate that this is an unconditional branch.
                instruction.declare_is_unconditional_jump = True

        # These condition codes indicate unconditionally/condition is not relevant
        if instruction.cs_insn.cc in (ARM_CC_AL, ARMCC_UNDEF):
            if instruction.id in (ARM_INS_B, ARM_INS_BL, ARM_INS_BLX, ARM_INS_BX, ARM_INS_BXJ):
                instruction.declare_is_unconditional_jump = True
            return InstructionCondition.UNDETERMINED

        value = self._read_register_name(instruction, self.flags_reg, emu)
        if value is None:
            # We can't reason about the value of flags register
            return InstructionCondition.UNDETERMINED

        N = (value >> 31) & 1
        Z = (value >> 30) & 1
        C = (value >> 29) & 1
        V = (value >> 28) & 1

        cc = {
            ARM_CC_EQ: Z,
            ARM_CC_NE: not Z,
            ARM_CC_HS: C,
            ARM_CC_LO: not C,
            ARM_CC_MI: N,
            ARM_CC_PL: not N,
            ARM_CC_VS: V,
            ARM_CC_VC: not V,
            ARM_CC_HI: C and not Z,
            ARM_CC_LS: Z or not C,
            ARM_CC_GE: N == V,
            ARM_CC_LT: N != V,
            ARM_CC_GT: not Z and (N == V),
            ARM_CC_LE: Z or (N != V),
        }.get(instruction.cs_insn.cc, None)

        if cc is None:
            return InstructionCondition.UNDETERMINED

        return InstructionCondition.TRUE if bool(cc) else InstructionCondition.FALSE

    @override
    def _resolve_target(self, instruction: PwndbgInstruction, emu: Emulator | None):
        target = super()._resolve_target(instruction, emu)
        if target is not None:
            # On interworking branches - branches that can enable Thumb mode - the target of a jump
            # has the least significant bit set to 1. This is not actually written to the PC
            # and instead the CPU puts it into the Thumb mode register bit.
            # This means we have to clear the least significant bit of the target.
            target = target & ~1

            if pwndbg.aglib.arch.name == "armcm" and target & 0xFF00_0000 == 0xFF00_0000:
                # If the top 8-bits of the return address are 0xFF, this indicates we are returning from an exception,
                # where the return address has been saved onto the stack
                return pwndbg.aglib.saved_register_frames.ARM_CORTEX_M_EXCEPTION_STACK.read_saved_register(
                    "pc"
                )

        return target

    # Currently not used
    def _memory_string_old(self, instruction: PwndbgInstruction, op: EnhancedOperand) -> str:
        parts = []

        if op.mem.base != 0:
            parts.append(instruction.cs_insn.reg_name(op.mem.base))

        if op.mem.disp != 0:
            parts.append(f"{op.mem.disp:#x}")

        if op.mem.index != 0:
            index = pwndbg.aglib.regs.read_reg(instruction.cs_insn.reg_name(op.mem.index))
            scale = op.mem.scale
            parts.append(f"{index}*{scale:#x}")

        return f"[{(', '.join(parts))}]"

    def read_thumb_bit(self, instruction: PwndbgInstruction, emu: Emulator) -> int | None:
        return 1 if instruction.cs_insn._cs._mode & CS_MODE_THUMB else 0

    @override
    def _immediate_string(self, instruction, operand):
        return "#" + super()._immediate_string(instruction, operand)

    @override
    def _read_register(
        self, instruction: PwndbgInstruction, operand_id: int, emu: Emulator
    ) -> int | None:
        # When `pc` is referenced in an operand (typically in a memory operand), the value it takes on
        # is `pc_at_instruction + 8`. In Thumb mode, you only add 4 to the instruction address.
        if operand_id == ARM_REG_PC:
            return instruction.address + (4 if self.read_thumb_bit(instruction, emu) else 8)

        return super()._read_register(instruction, operand_id, emu)

    @override
    def _parse_memory(
        self, instruction: PwndbgInstruction, op: EnhancedOperand, emu: Emulator
    ) -> int | None:
        """
        Parse the `ArmOpMem` Capstone object to determine the concrete memory address used.

        Types of memory operands:
            [Rn]
            [Rn, #imm]
            [Rn, Rm]
            [Rn, Rm, <shift> #imm]

        Capstone represents the object a bit differently then AArch64 to align with the underlying architecture of Arm.

        This representation will change in Capstone 6:
            https://github.com/capstone-engine/capstone/issues/2281
            https://github.com/capstone-engine/capstone/pull/1949
        """

        target = 0

        # All memory operands have `base` defined
        base = self._read_register(instruction, op.mem.base, emu)
        if base is None:
            return None

        if op.mem.base == ARM_REG_PC:
            # The PC as the base register is a special case - it will align the address to a word (32-bit) boundary
            # Explanation: https://stackoverflow.com/a/29588678
            # See "Operation" at the bottom of https://developer.arm.com/documentation/ddi0597/2024-03/Base-Instructions/LDR--literal---Load-Register--literal--
            base = align_down(4, base)

        target = base

        # On post index, the base pointer is incremented after the memory dereference
        if not instruction.cs_insn.post_index:
            target += op.mem.disp * (-1 if op.cs_op.subtracted else 1)

        # If there is an index register
        if op.mem.index != 0:
            index = self._read_register(instruction, op.mem.index, emu)
            if index is None:
                return None

            # Optionally apply shift to the index register
            if op.cs_op.shift.type != 0:
                index = ARM_BIT_SHIFT_MAP[op.cs_op.shift.type](index, op.cs_op.shift.value, 32)

            target += index * op.mem.scale

        return target

    @override
    def _parse_register(
        self, instruction: PwndbgInstruction, op: EnhancedOperand, emu: Emulator
    ) -> int | None:
        """
        Register operands can have optional shifts in Arm
        """
        target = super()._parse_register(instruction, op, emu)
        if target is None:
            return None

        # We need this to retain the value of the un-shifted register in some annotations, such as shifts
        op.before_value_no_modifiers = target

        # Optionally apply shift to the index register
        if op.cs_op.shift.type != 0:
            target = ARM_BIT_SHIFT_MAP.get(op.cs_op.shift.type, lambda *a: None)(
                target, op.cs_op.shift.value, 32
            )

        return target



from __future__ import annotations

import argparse

import pwndbg.aglib.file
import pwndbg.commands
import pwndbg.wrappers.checksec
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="Prints out the binary security settings using `checksec`.",
)

parser.add_argument("-f", "--file", type=str, help="Specify the file to run `checksec` on.")


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MISC,
    examples="""
pwndbg> checksec
File:     /usr/bin/bash
Arch:     amd64
RELRO:      Full RELRO
Stack:      Canary found
NX:         NX enabled
PIE:        PIE enabled
SHSTK:      Enabled
IBT:        Enabled
""",
)
@pwndbg.commands.OnlyWithFile
def checksec(file: str) -> None:
    local_path = file or pwndbg.aglib.file.get_proc_exe_file()
    print(pwndbg.wrappers.checksec.get_raw_out(local_path))



"""
Attempt to describe the data at a memory address as best as possible.

Only when the data is not a pointer.

Currently prints out code, integers, or strings, in a best-effort manner
dependent on page permissions, the contents of the data, and any
supplemental information sources (e.g. active IDA Pro connection).
"""

from __future__ import annotations

import string

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.disasm.disassembly
import pwndbg.aglib.memory
import pwndbg.aglib.strings
import pwndbg.aglib.vmmap
import pwndbg.color.enhance as E
import pwndbg.color.memory
import pwndbg.dintegration
import pwndbg.lib.pretty_print
from pwndbg import color


def int_str(value: int) -> str:
    retval = pwndbg.lib.pretty_print.int_to_string(value)

    # Try to unpack the value as a string
    packed = pwndbg.aglib.arch.pack(value)
    if all(c in string.printable.encode("utf-8") for c in packed):
        if len(retval) > 4:
            retval = "{} ({!r})".format(retval, str(packed.decode("ascii", "ignore")))

    return retval


# @pwndbg.lib.cache.cache_until("stop")
def enhance(
    value: int,
    code: bool = True,
    safe_linking: bool = False,
    attempt_dereference=True,
    enhance_string_len: int = None,
) -> str:
    """
    Given the last pointer in a chain, attempt to characterize

    Note that 'the last pointer in a chain' may not at all actually be a pointer.

    Additionally, optimizations are made based on various sources of data for
    'value'. For example, if it is set to RWX, we try to get information on whether
    it resides on the stack, or in a RW section that *happens* to be RWX, to
    determine which order to print the fields.

    Arguments:
        value: Value to enhance
        code: Hint that indicates the value may be an instruction
        safe_linking: Whether this chain use safe-linking
        enhance_string_len: The length of string to display for enhancement of the last pointer
    """
    value = int(value)

    page = pwndbg.aglib.vmmap.find(value)

    # If it's not in a page we know about, try to dereference
    # it anyway just to test.
    can_read = True
    if not attempt_dereference or not page or None is pwndbg.aglib.memory.peek(value):
        can_read = False

    # If it's a pointer that we told we cannot deference, then color it accordingly and add symbol if can
    if page and not attempt_dereference:
        return pwndbg.color.memory.get_address_and_symbol(
            value, pwndbg.dintegration.manager.get_stack_var_dict_all()
        )

    if not can_read:
        return E.integer(int_str(value))

    # It's mapped memory, or we can at least read it.
    # Try to find out if it's a string.
    instr: str | None = None
    exe = page and page.execute
    rwx = page and page.rwx

    # For the purpose of following pointers, don't display
    # anything on the stack or heap as 'code'
    if "[stack" in page.objfile or "[heap" in page.objfile:
        rwx = exe = False

    if exe:
        pwndbg_instr = pwndbg.aglib.disasm.disassembly.one_raw(value)
        if pwndbg_instr:
            pwndbg.aglib.disasm.arch.basic_enhance(pwndbg_instr)
            # For telescoping, we don't want the extra spaces between the mnemonic and operands
            # which are baked in during enhancement. This removes those spaces.
            instr = " ".join(pwndbg_instr.asm_string.split())

    szval = pwndbg.aglib.strings.get(value, maxlen=enhance_string_len) or None
    szval0 = szval
    if szval:
        szval = E.string(repr(szval))

    # Fix for case when we can't read the end address anyway (#946)
    if value + pwndbg.aglib.arch.ptrsize > page.end:
        return E.integer(int_str(value))

    intval = pwndbg.aglib.memory.read_pointer_width(value)
    if safe_linking:
        intval ^= value >> 12
    intval0 = intval
    intval = E.integer(pwndbg.lib.pretty_print.int_to_string(intval & pwndbg.aglib.arch.ptrmask))

    retval = []

    # print([instr,intval0,szval])
    if not code:
        instr = None

    # If it's on the stack, don't display it as code in a chain.
    if instr and "[stack" in page.objfile:
        retval = [intval, szval]

    # If it's RWX but a small value, don't display it as code in a chain.
    elif instr and rwx and intval0 < 0x1000:
        retval = [intval, szval]

    # If it's an instruction and *not* RWX, display it unconditionally
    elif instr and exe:
        if not rwx:
            if szval:
                retval = [instr, szval]
            else:
                retval = [instr]
        else:
            retval = [instr, intval, szval]

    # Otherwise strings have preference
    elif szval:
        if len(szval0) < pwndbg.aglib.arch.ptrsize:
            retval = [intval, szval]
        else:
            retval = [szval]

    # And then integer
    else:
        # It might be a pointer or just a plain integer
        new_page = pwndbg.aglib.vmmap.find(intval0)
        if new_page:
            return pwndbg.color.memory.get_address_and_symbol(
                intval0, pwndbg.dintegration.manager.get_stack_var_dict_all()
            )
        return E.integer(int_str(intval0))

    retval = tuple(filter(lambda x: x is not None, retval))

    if len(retval) == 0:
        return E.unknown("???")

    if len(retval) == 1:
        return retval[0]

    return retval[0] + E.comment(color.strip(f" /* {'; '.join(retval[1:])} */"))



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from capstone6pwndbg import *  # noqa: F403
from capstone6pwndbg.loongarch import *  # noqa: F403
from typing_extensions import override

import pwndbg.aglib.disasm.arch
import pwndbg.lib.disasm.helpers as bit_math
from pwndbg.aglib.disasm.instruction import InstructionCondition
from pwndbg.aglib.disasm.instruction import PwndbgInstruction

# Emulator currently requires GDB, and we only use it here for type checking.
if TYPE_CHECKING:
    from pwndbg.emu.emulator import Emulator

CONDITION_RESOLVERS: dict[int, Callable[[list[int]], bool]] = {
    LOONGARCH_INS_BEQZ: lambda ops: ops[0] == 0,
    LOONGARCH_INS_BNEZ: lambda ops: ops[0] != 0,
    LOONGARCH_INS_BEQ: lambda ops: ops[0] == ops[1],
    LOONGARCH_INS_BNE: lambda ops: ops[0] != ops[1],
    LOONGARCH_INS_BGE: lambda ops: bit_math.to_signed(ops[0], pwndbg.aglib.arch.ptrbits)
    >= bit_math.to_signed(ops[1], pwndbg.aglib.arch.ptrbits),
    LOONGARCH_INS_BLT: lambda ops: bit_math.to_signed(ops[0], pwndbg.aglib.arch.ptrbits)
    < bit_math.to_signed(ops[1], pwndbg.aglib.arch.ptrbits),
    LOONGARCH_INS_BLTU: lambda ops: ops[0] < ops[1],
    LOONGARCH_INS_BGEU: lambda ops: ops[0] >= ops[1],
}


LOONGARCH_LOAD_INSTRUCTIONS: dict[int, int] = {}

LOONGARCH_STORE_INSTRUCTIONS: dict[int, int] = {}

LOONGARCH_BINARY_OPERATIONS: dict[int, str] = {}


# This class enhances 64-bit Loongarch
class Loong64DisassemblyAssistant(pwndbg.aglib.disasm.arch.DisassemblyAssistant):
    def __init__(self, architecture) -> None:
        super().__init__(architecture)

        self.annotation_handlers: dict[int, Callable[[PwndbgInstruction, Emulator], None]] = {}

    @override
    def _condition(self, instruction: PwndbgInstruction, emu: Emulator) -> InstructionCondition:
        if len(instruction.operands) == 0:
            return InstructionCondition.UNDETERMINED

        # Not using list comprehension because they run in a separate scope in which super() does not exist
        resolved_operands: list[int] = []
        for op in instruction.operands:
            resolved_operands.append(
                super()._resolve_used_value(op.before_value, instruction, op, emu)
            )

        # If any of the relevent operands are None (we can't reason about them), quit.
        if any(value is None for value in resolved_operands[:-1]):
            # Note the [:-1]. Loongarch jump instructions have the target as the last operand
            # https://loongson.github.io/LoongArch-Documentation/LoongArch-Vol1-EN.html#_beqz_bnez
            return InstructionCondition.UNDETERMINED

        conditional = CONDITION_RESOLVERS.get(instruction.id, lambda *a: None)(resolved_operands)

        if conditional is None:
            return InstructionCondition.UNDETERMINED

        return InstructionCondition.TRUE if conditional else InstructionCondition.FALSE

    @override
    def _resolve_target(self, instruction: PwndbgInstruction, emu: Emulator | None):
        if instruction.id == LOONGARCH_INS_ALIAS_RET:
            return self._read_register_name(instruction, "ra", emu)

        if instruction.id == LOONGARCH_INS_JIRL:
            if (offset_reg := instruction.operands[1].before_value) is not None:
                return offset_reg + (instruction.operands[2].before_value << 2)

        return super()._resolve_target(instruction, emu)



from __future__ import annotations

import shlex
import string

import pwndbg.aglib.file
import pwndbg.aglib.proc
import pwndbg.aglib.qemu
import pwndbg.auxv
import pwndbg.commands
import pwndbg.lib.cache
import pwndbg.lib.net
from pwndbg.color import message
from pwndbg.commands import CommandCategory

"""
PEDA prints it out like this:

exe = /bin/bash
fd[0] -> /dev/pts/96
fd[1] -> /dev/pts/96
fd[2] -> /dev/pts/96
pid = 31102
ppid = 31096
uid = [287138, 287138, 287138, 287138]
gid = [5000, 5000, 5000, 5000]
"""

capabilities = {
    0: "CAP_CHOWN",
    1: "CAP_DAC_OVERRIDE",
    2: "CAP_DAC_READ_SEARCH",
    3: "CAP_FOWNER",
    4: "CAP_FSETID",
    5: "CAP_KILL",
    6: "CAP_SETGID",
    7: "CAP_SETUID",
    8: "CAP_SETPCAP",
    9: "CAP_LINUX_IMMUTABLE",
    10: "CAP_NET_BIND_SERVICE",
    11: "CAP_NET_BROADCAST",
    12: "CAP_NET_ADMIN",
    13: "CAP_NET_RAW",
    14: "CAP_IPC_LOCK",
    15: "CAP_IPC_OWNER",
    16: "CAP_SYS_MODULE",
    17: "CAP_SYS_RAWIO",
    18: "CAP_SYS_CHROOT",
    19: "CAP_SYS_PTRACE",
    20: "CAP_SYS_PACCT",
    21: "CAP_SYS_ADMIN",
    22: "CAP_SYS_BOOT",
    23: "CAP_SYS_NICE",
    24: "CAP_SYS_RESOURCE",
    25: "CAP_SYS_TIME",
    26: "CAP_SYS_TTY_CONFIG",
    27: "CAP_MKNOD",
    28: "CAP_LEASE",
    29: "CAP_AUDIT_WRITE",
    30: "CAP_AUDIT_CONTROL",
    31: "CAP_SETFCAP",
    32: "CAP_MAC_OVERRIDE",
    33: "CAP_MAC_ADMIN",
    34: "CAP_SYSLOG",
    35: "CAP_WAKE_ALARM",
    36: "CAP_BLOCK_SUSPEND",
}


def tcp(tid: int):
    # For reference, see:
    # https://www.kernel.org/doc/Documentation/networking/proc_net_tcp.txt
    """
    It will first list all listening TCP sockets, and next list all established
    TCP connections. A typical entry of /proc/net/tcp would look like this (split
    up into 3 parts because of the length of the line):
    """
    data = pwndbg.aglib.file.get(f"/proc/{tid}/net/tcp").decode()
    return pwndbg.lib.net.tcp(data, pwndbg.aglib.arch.endian)


def tcp6(tid: int):
    data = pwndbg.aglib.file.get(f"/proc/{tid}/net/tcp6").decode()
    return pwndbg.lib.net.tcp6(data, pwndbg.aglib.arch.endian)


def unix(tid: int):
    # We use errors=ignore because of https://github.com/pwndbg/pwndbg/issues/1544
    # TODO/FIXME: this may not be the best solution because we may end up with
    # invalid UDS data. Can this be a problem?
    data = pwndbg.aglib.file.get(f"/proc/{tid}/net/unix").decode(errors="ignore")
    return pwndbg.lib.net.unix(data)


def netlink(tid: int):
    data = pwndbg.aglib.file.get(f"/proc/{tid}/net/netlink").decode()
    return pwndbg.lib.net.netlink(data)


class Process:
    def __init__(self, pid=None, tid=None) -> None:
        if pid is None:
            pid = pwndbg.aglib.proc.pid()
        if tid is None:
            tid = pwndbg.aglib.proc.tid()
        if not tid:
            tid = pid
        self.pid = pid
        self.tid = tid

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def selinux(self) -> str:
        path = f"/proc/{self.pid}/task/{self.tid}/attr/current"
        try:
            raw = pwndbg.aglib.file.get(path)
            return raw.decode().rstrip("\x00").strip()
        except OSError:
            # This file sometimes cannot be read. This indicates SELinux is not enabled
            return ""

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def cmdline(self):
        raw = pwndbg.aglib.file.get(f"/proc/{self.pid}/cmdline")
        return " ".join(map(shlex.quote, raw.decode().split("\x00")))

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def cwd(self) -> str:
        link = pwndbg.aglib.file.readlink(f"/proc/{self.pid}/cwd")
        return f"'{link}'"

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def status(self):
        raw = pwndbg.aglib.file.get(f"/proc/{self.pid}/task/{self.pid}/status")

        status = {}
        for line in raw.splitlines():
            if not line:
                continue

            k_v = line.split(maxsplit=1)

            if len(k_v) == 1:
                k_v.append(b"")

            k, v = k_v

            k = k.decode("latin-1")
            v = v.decode("latin-1")

            k = k.lower().rstrip(":")

            # bit fields
            if set(v) < set(string.hexdigits) and len(v) == 16:
                try:
                    v = int(v, 16)
                except AttributeError:
                    pass

            # vm stats
            elif v.endswith(" kB"):
                v = int(v.split()[0]) * (1 << 10)
            elif v.endswith(" mB"):
                v = int(v.split()[0]) * (1 << 20)

            # misc integers like pid and ppid
            elif str(v).isdigit():
                v = int(v)

            # uid and gid and groups
            elif all(s.isdigit() for s in v.split()):
                v = list(map(int, v.split()))

            # capability sets
            if k in ["capeff", "capinh", "capprm", "capbnd"]:
                orig: int = v
                v = []
                for i in range(max(capabilities) + 1):
                    if (orig >> i) & 1 == 1:
                        v.append(capabilities[i])

            status[k] = v
            setattr(self, k, v)
        return status

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def open_files(self):
        fds = {}

        for i in range(self.fdsize):
            link = pwndbg.aglib.file.readlink(f"/proc/{pwndbg.aglib.proc.pid()}/fd/{i}")

            if link:
                fds[i] = link

        return fds

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def connections(self):
        # Connections look something like this:
        # socket:[102422]
        fds = self.open_files
        socket = "socket:["
        result = []

        functions = [tcp, tcp6, unix, netlink]

        for fd, path in fds.items():
            if socket not in path:
                continue

            inode = path[len(socket) : -1]
            inode = int(inode)

            for func in functions:
                for x in func(self.tid):
                    if x.inode == inode:
                        x.fd = fd
                        result.append(x)

        return tuple(result)


@pwndbg.commands.Command("Gets the pid.", aliases=["getpid"], category=CommandCategory.PROCESS)
@pwndbg.commands.OnlyWhenRunning
def pid() -> None:
    print(pwndbg.aglib.proc.pid())


@pwndbg.commands.Command(
    "Display information about the running process.", category=CommandCategory.PROCESS
)
@pwndbg.commands.OnlyWhenRunning
def procinfo() -> None:
    """
    Display information about the running process.
    """
    if pwndbg.aglib.qemu.is_qemu():
        print(
            message.error(
                "QEMU target detected: showing result for the qemu process"
                " - so it will be a bit inaccurate (excessive for the parts"
                " used directly by the qemu process)"
            )
        )
    exe = pwndbg.auxv.get().AT_EXECFN
    print(f"{'exe':<10} {exe!r}")

    proc = Process()

    # qemu-usermode fail!
    if not proc.status:
        return

    print(f"{'cmdline':<10} {proc.cmdline}")
    print(f"{'cwd':<10} {proc.cwd}")

    files = dict(proc.open_files)

    for c in proc.connections:
        files[c.fd] = str(c)

    print(f"{'pid':<10} {proc.pid}")
    print(f"{'tid':<10} {proc.tid}")

    if proc.selinux and proc.selinux != "unconfined":
        print(f"{'selinux':<10} {proc.selinux}")

    print(f"{'ppid':<10} {proc.ppid}")
    print(f"{'uid':<10} {proc.uid}")
    print(f"{'gid':<10} {proc.gid}")
    print(f"{'groups':<10} {proc.groups}")

    for fd, path in files.items():
        if not set(path) < set(string.printable):
            path = repr(path)
        print(f"{f'fd[{fd}]':<10} {path}")

    return



from __future__ import annotations

import argparse

import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.chain
import pwndbg.commands
import pwndbg.dbg_mod
from pwndbg.color import message
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="""Dumps the elements of a linked list.

This command traverses the linked list beginning at a given element, dumping its
contents and the contents of all the elements that come after it in the list.
Traversal is configurable and can handle multiple types of linked lists, but will
always stop when a cycle is detected.

The path to the first element can be any GDB expression that evaluates to either
the first element directly, or a to pointer to it. The next element is the name
of the field containing the next pointer, in either the structure itself or in
the structure given by --inner.

An address value may be given with --sentinel that signals the end of the list.
By default, the value used is NULL (0).

If only one field inside each node is desired, it can be printed exclusively by
specifying its name with --field.

This command supports traversing three types of linked lists, classified by how
the next pointer can be found in the structure and what type it is:
    1 - Next pointer is field of structure, type is the same as structure.
    2 - Next pointer is field of inner nested structure, pointed to type is the
        same as outer structure.
    3 - Next pointer is field of inner nested structure, pointed to type is the
        same as inner structure.
Types 2 and 3 require --inner to be specified.

Example 1:

```
struct node {
    int value;
    struct node *next;
};
struct node node_c = { 2, NULL };
struct node node_b = { 1, &node_c };
struct node node_a = { 0, &node_b };
```

pwndbg> plist node_a next
0x4000011050 <node_a>: {
  value = 0,
  next = 0x4000011040 <node_b>
}
0x4000011040 <node_b>: {
  value = 1,
  next = 0x4000011010 <node_c>
}
0x4000011010 <node_c>: {
  value = 2,
  next = 0x0
}

Example 2:

```
struct node_inner_a {
    struct node_inner_a *next;
};
struct inner_a_node {
    int value;
    struct node_inner_a inner;
};
struct inner_a_node inner_a_node_c = { 2, { NULL } };
struct inner_a_node inner_a_node_b = { 1, { &inner_a_node_c.inner } };
struct inner_a_node inner_a_node_a = { 0, { &inner_a_node_b.inner } };
```

pwndbg> plist inner_a_node_a -i inner next
0x4000011070 <inner_a_node_a>: {
  value = 0,
  inner = {
    next = 0x4000011068 <inner_a_node_b+8>
  }
}
0x4000011060 <inner_a_node_b>: {
  value = 1,
  inner = {
    next = 0x4000011028 <inner_a_node_c+8>
  }
}
0x4000011020 <inner_a_node_c>: {
  value = 2,
  inner = {
    next = 0x0
  }
}

Example 3:

```
struct inner_b_node;
struct node_inner_b {
    struct inner_b_node *next;
};
struct inner_b_node {
    int value;
    struct node_inner_b inner;
};
struct inner_b_node inner_b_node_c = { 2, { NULL } };
struct inner_b_node inner_b_node_b = { 1, { &inner_b_node_c } };
struct inner_b_node inner_b_node_a = { 0, { &inner_b_node_b } };
```

pwndbg> plist inner_b_node_a -i inner next
0x4000011090 <inner_b_node_a>: {
  value = 0,
  inner = {
    next = 0x4000011080 <inner_b_node_b>
  }
}
0x4000011080 <inner_b_node_b>: {
  value = 1,
  inner = {
    next = 0x4000011030 <inner_b_node_c>
  }
}
0x4000011030 <inner_b_node_c>: {
  value = 2,
  inner = {
    next = 0x0
  }
}

""",
)
parser.add_argument(
    "path",
    type=str,
    help="The first element of the linked list",
)
parser.add_argument(
    "next", type=str, help="The name of the field pointing to the next element in the list"
)
parser.add_argument(
    "-s",
    "--sentinel",
    dest="sentinel",
    type=int,
    default=0,
    help="The address that stands in for an end of list marker in a non-cyclic list",
)
parser.add_argument(
    "-i",
    "--inner",
    dest="inner_name",
    type=str,
    help="The name of the inner nested structure where the next pointer is stored",
)
parser.add_argument(
    "-f",
    "--field",
    dest="field_name",
    type=str,
    help="The name of the field to be displayed, if only one is desired",
)
parser.add_argument(
    "-o",
    "--offset",
    dest="offset",
    type=int,
    default=0,
    help="The offset of the first list element to display. Defaults to zero.",
)
parser.add_argument(
    "-c",
    "--count",
    dest="count",
    type=int,
    default=None,
    help="The number of elements to display. Defaults to the value of dereference-limit.",
)


@pwndbg.commands.Command(parser, command_name="plist", category=CommandCategory.MISC)
def plist(
    path: str,
    next: str,
    sentinel: int,
    inner_name: str | None,
    field_name: str | None,
    offset: int,
    count: int | None = None,
) -> None:
    # Have GDB parse the path for us and check if it's valid.
    try:
        first = pwndbg.dbg.selected_frame().evaluate_expression(path)
    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"{e}"))
        return

    if first.is_optimized_out:
        print(message.error(f"{path} has been optimized out"))
        return

    if count is None:
        count = pwndbg.config.dereference_limit
    if count <= 0:
        print("count <= 0: not displaying any elements")
        return

    # We suport being passed either a pointer to the first structure or the
    # structure itself, for the sake of convenience. But we don't bother with
    # chains of pointers. Additionally, we pick the correct separator to use
    # for the error mesages.
    sep = "."
    deref = ""
    if first.type.code == pwndbg.dbg_mod.TypeCode.POINTER:
        sep = "->"
        deref = "*"
        try:
            first = first.dereference()
        except pwndbg.dbg_mod.Error as e:
            print(message.error(f"Pointer at {path} could not be dereferenced: {e}"))
            return

    if first.type.code == pwndbg.dbg_mod.TypeCode.POINTER:
        print(message.error(f"{path} is not a value or a single pointer to one"))
        return

    if first.address is None:
        print(message.error(f"{deref}{path} is not addressable"))
        return

    if first.is_optimized_out:
        print(message.error(f"{deref}{path} has been optimized out"))
        return

    # If there is an inner element we have to use, find it.
    inner = None
    inner_sep = ""
    if inner_name is not None:
        try:
            inner = first[inner_name]
            inner_sep = "->"
        except pwndbg.dbg_mod.Error as e:
            print(message.error(f"Cannot find component {inner_name} in {path}: {e}"))
            return
        if inner.is_optimized_out:
            print(message.error(f"{path}{sep}{inner_name} has been optimized out"))
            return

    # Resolve the pointer to the next structure, wherever it may be, and make
    # sure that we can use it to traverse the linked list.
    next_ptr_loc = first
    next_ptr_name = next
    try:
        if inner is None:
            next_ptr = first[next]
        else:
            next_ptr = inner[next]
            next_ptr_loc = inner
            next_ptr_name = f"{inner_name}.{next}"
    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"Cannot find component {next_ptr_name} in {path}: {e}"))
        return

    if next_ptr.is_optimized_out:
        print(message.error(f"{path}{sep}{next_ptr_name} has been optimized out"))
        return

    # Check if the field is a pointer type, or an integer type with pointer size
    # (e.g., size_t, uintptr_t, unsigned long on 64-bit systems)
    # Strip typedefs to get the underlying type (e.g., size_t -> unsigned long)
    underlying_type = next_ptr.type.strip_typedefs()
    is_pointer_type = underlying_type.code == pwndbg.dbg_mod.TypeCode.POINTER
    is_pointer_sized_int = (
        underlying_type.code == pwndbg.dbg_mod.TypeCode.INT
        and underlying_type.sizeof == pwndbg.aglib.arch.ptrsize
    )

    if not (is_pointer_type or is_pointer_sized_int):
        print(
            message.error(f"{path}{sep}{next_ptr_name} is not a pointer or pointer-sized integer")
        )
        return

    # If the user wants a specific field to be displayed, resolve it.
    field_offset = None
    field_type = None
    if field_name is not None:
        try:
            field = first[field_name]
        except pwndbg.dbg_mod.Error as e:
            print(message.error(f"Cannot find component {field_name} in {path}: {e}"))
            return
        field_type = field.type

        bit_offset = bit_offset_of_field(first, field_name)
        if bit_offset is None:
            print(
                message.error(
                    f"{path}{sep}{field_name} has no known offset \
                from {deref}{path}"
                )
            )
            return
        byte_offset = get_byte_offset(bit_offset)
        if byte_offset is None:
            print(
                message.error(
                    f"{field_name} is a non-whole number of bytes \
                {bit_offset} bits) offset from {deref}{path}"
                )
            )
            return
        field_offset = byte_offset

    # Figure out the offset of the inner structure, if any, in typeof(first).
    inner_offset = 0
    if inner is not None:
        bit_offset = bit_offset_of_field(first, inner_name)
        if bit_offset is None:
            print(
                message.error(
                    f"{path}{sep}{inner_name} has no known offset \
                from {deref}{path}"
                )
            )
            return
        byte_offset = get_byte_offset(bit_offset)
        if byte_offset is None:
            print(
                message.error(
                    f"{inner_name} is a non-whole number of bytes \
                {bit_offset} bits) offset from {deref}{path}"
                )
            )
            return
        inner_offset = byte_offset

    # Figure out the offset of the next pointer in its containing type.
    bit_offset = bit_offset_of_field(next_ptr_loc, next)
    if bit_offset is None:
        print(
            message.error(
                f"{path}{sep}{next_ptr_name} has no known offset \
            from {deref}{path}{inner_sep}{inner_name}"
            )
        )
        return
    byte_offset = get_byte_offset(bit_offset)
    if byte_offset is None:
        print(
            message.error(
                f"{path}{sep}{next_ptr_name} is a non-whole \
            number of bytes {bit_offset} bits) offset from \
            {deref}{path}{inner_sep}{inner_name}"
            )
        )
        return
    next_offset = byte_offset

    # If the next pointer points to an intance of the inner structure, we will
    # additionally have to do the equivalent of container_of(next, typeof(first),
    # inner_name).
    #
    # Here, we figure out how many bytes to subtract from *typeof(inner) so that
    # we can have a *typeof(first).
    pointee_offset = 0

    # For pointer-sized integers (like size_t), we can't validate the target type
    # since they don't have a target() method. We assume they point to the outer structure.
    if is_pointer_sized_int:
        # For integer types, we assume pointee_offset is 0 (pointing to outer structure)
        # unless we have an inner structure, in which case the user needs to use --inner
        # and we can't automatically determine the offset.
        if inner is not None:
            # We can't determine if it points to inner or outer type, so assume outer
            pointee_offset = 0
    elif inner is not None and next_ptr.type.target() == inner.type:
        pointee_offset = inner_offset
    elif next_ptr.type.target() == first.type:
        # We've already got everything we need for this mode.
        pass
    else:
        print(
            message.error(
                f"{deref}{path}{sep}{next_ptr_name} has a \
            different type than {path}"
            )
        )
        return

    # Now, we follow the chain. We have to do this in two steps, as, because we
    # have the address of the first outer structure, in case pointee_offset is
    # not zero, the offset for the first element will not be the same as the one
    # for all of the elements after it.
    offset0 = inner_offset + next_offset
    offset1 = offset0 - pointee_offset

    total = offset + count

    if total >= 2:
        addresses = pwndbg.chain.get(int(first.address), limit=1, offset=offset0)
        if len(addresses) > 1 and total >= 3:
            addresses.extend(
                pwndbg.chain.get(addresses[1], offset=offset1, include_start=False, limit=total - 2)
            )
    else:
        addresses = [int(first.address)]

    # Finally, dump the information in the addresses we've just gathered.
    for i, address in enumerate(addresses):
        if i < offset:
            continue
        if address in (0, sentinel):
            break
        try:
            # Always make sure we have the address of the outer structure.
            if i > 0:
                address -= pointee_offset

            # Read the data and print it out.
            target_type = first.type
            target_address = address
            if field_offset is not None:
                target_type = field_type
                target_address = address + field_offset

            value = pwndbg.aglib.memory.get_typed_pointer_value(target_type, target_address)

            symbol = pwndbg.aglib.symbol.resolve_addr(target_address)
            symbol = f"<{symbol}>" if symbol else ""

            print(f"{target_address:#x} {symbol}: {value.value_to_human_readable()}")
        except pwndbg.dbg_mod.Error as e:
            print(message.error(f"Cannot dereference {address:#x} for list link #{i + 1}: {e}"))
            print(message.error("Is the linked list corrupted or is the sentinel value wrong?"))
            return


# Helper functions to resolve the offsets of fields inside a structure.
def bit_offset_of_field(struct, field_name, inner_name=None):
    offset = None
    for field in struct.type.fields():
        if field.name == field_name:
            offset = field.bitpos
    return offset


def get_byte_offset(bit_offset):
    if bit_offset % 8 != 0:
        return None
    return bit_offset // 8



from __future__ import annotations

import gdb

import pwndbg.lib.cache


class Breakpoint(gdb.Breakpoint):
    """
    Breakpoint class, similar to gdb.Breakpoint, but clears the caches
    associated with the stop event before determining whether it should stop the
    inferior or not.

    Unlike gdb.Breakpoint, users of this class should override `should_stop()`,
    instead of `stop()`, as the latter is used to do cache invalidation.
    """

    def stop(self) -> bool:
        # Clear the cache for the stop event.
        pwndbg.lib.cache.clear_cache(pwndbg.lib.cache.CacheUntilEvent.STOP)
        return self.should_stop()

    def should_stop(self) -> bool:
        """
        This function is called whenever this breakpoint is hit in the code and
        its return value determines whether the inferior will be stopped.
        """
        return True



from __future__ import annotations

import sys


def system_decode(b: bytes) -> str:
    """
    LLDB requires Python strings in many places where it makes sense to accept
    bytes values. This is mostly an artifact of how Swig maps C `char*` to
    `str` in Python, but since Swig will refuse bytes objects, we have to figure
    out a way to pass this data as a regular string object, even if that's
    nonsensical in Python terms.

    This function tries its best to resolve that by decoding it with the same
    decoder the filesystem uses, and, failing that, ASCII.
    """

    native = sys.getfilesystemencoding()

    try:
        return b.decode(native)
    except UnicodeDecodeError:
        pass

    return b.decode("ascii")



"""
Pwndbg profiling.
"""

from __future__ import annotations

import cProfile
import time

profiler: Profiler | None = None


def init(p: cProfile.Profile, _start_time: float | None) -> None:
    global profiler
    profiler = Profiler(p)
    profiler._start_time = _start_time


class Profiler:
    def __init__(self, p: cProfile.Profile) -> None:
        self._profiler = p
        self._start_time: float | None = None

    def print_time_elapsed(self) -> None:
        assert self._start_time is not None
        print("Time Elapsed:", time.time() - self._start_time)

    def start(self) -> None:
        self._start_time = time.time()
        self._profiler.enable()

    def stop(self, filename: str | None = None) -> None:
        if not filename:
            filename = f"pwndbg-{int(time.time())}.pstats"

        self.print_time_elapsed()
        self._profiler.disable()
        self._start_time = None

        self._profiler.dump_stats(filename)
